"""
Frontend for exposing TensorFlow-specific functionality.
"""

from __future__ import annotations
from typing import Union, Sequence
import tensorflow as tf

from returnn.util.basic import NotSpecified
from returnn.frontend_api import Frontend, RawTensorTypes
from returnn.tensor import Tensor, Dim
from returnn.tf.util import basic as tf_util
from ._internal_frontend import TFInternalFrontend

_TT = Tensor[tf.Tensor]


# Ignore this warning until we really expect that we implemented everything.
# noinspection PyAbstractClass
class TFFrontend(Frontend[tf.Tensor]):
    """
    TensorFlow low-level frontend, operating on tf.Tensor
    """

    RawTensorType = tf.Tensor
    is_tensorflow = True
    _internal_frontend = TFInternalFrontend

    @staticmethod
    def convert_to_tensor(value: Union[_TT, tf.Tensor, RawTensorTypes]) -> _TT:
        """
        :param value:
        :return: tensor
        """
        if isinstance(value, Tensor):
            return value
        value = tf.convert_to_tensor(value)
        assert isinstance(value, tf.Tensor)
        assert value.shape.as_list() == [], f"scalar expected, got {value}"
        return Tensor("const", raw_tensor=value, dims=[], dtype=value.dtype.base_dtype.name)

    @staticmethod
    def range_over_dim(dim: Dim) -> _TT:
        """
        :param dim:
        :return: range over dim
        """
        out = Tensor(name=dim.description or "range_over_dim", dims=[dim], sparse_dim=dim)
        dim_value = dim.get_dim_value()
        out.raw_tensor = tf.range(0, dim_value, dtype=out.dtype)
        return out

    @staticmethod
    def reduce(source: _TT, *, mode: str, axis: Union[Dim, Sequence[Dim]], use_time_mask: bool = NotSpecified) -> _TT:
        """Reduce"""
        x = source
        axes = x.get_axes_from_description(axis)
        if use_time_mask in (None, NotSpecified):
            use_time_mask = any(x.has_dynamic_size(a) for a in axes)
        out_data = x.copy_template()
        dim_tags = [dim_tag for i, dim_tag in enumerate(x.dim_tags) if i not in axes]
        out_data = out_data.copy_template_new_dim_tags(dim_tags)
        sparse_out = mode.lower().startswith("arg")
        if sparse_out:
            assert len(axes) == 1
            out_data.sparse_dim = x.dim_tags[axes[0]]
            out_data.dtype = "int32"
        assert isinstance(use_time_mask, bool)
        mode = mode.lower()
        if mode == "avg":  # alias
            mode = "mean"
        reduce_abs_funcs = {
            name: getattr(tf, "reduce_%s" % name) for name in ["max", "min", "sum", "logsumexp", "any", "all"]
        }
        reduce_rel_func = {"mean": tf.reduce_mean}
        arg_funcs = {name: getattr(tf, name) for name in ["argmax", "argmin"]}
        funcs = dict(list(reduce_abs_funcs.items()) + list(reduce_rel_func.items()) + list(arg_funcs.items()))
        assert mode in funcs, "invalid mode %r. choose from: %r" % (mode, funcs)
        f = funcs[mode]
        x_ = x.placeholder
        # Check if we should ignore some frames, e.g. via masking.
        correction_factor = None
        if use_time_mask and any(x.has_dynamic_size(a) for a in axes):
            if x.batch_dim_axis in axes and x.time_dim_axis in axes and len(axes) == 2:
                assert mode not in arg_funcs, "unexpected arg reduce for multiple axes"
                # Flattening.
                axes = [a if (a < x.time_dim_axis) else (a - 1) for a in axes if a != x.time_dim_axis]
                x = x.copy_time_flattened()
                x_ = x.placeholder

            else:
                # Fhe fastest and simplest way is masking.
                for axis in axes:
                    if axis == x.batch_dim_axis:
                        continue
                    if not x.has_dynamic_size(axis):
                        continue
                    mask = x.get_sequence_mask_broadcast(axis=axis)

                    zeros = tf.zeros((), dtype=x.placeholder.dtype)
                    # Cannot call x.placeholder.dtype.{min,max} in case input is e.g. a bool
                    if x.placeholder.dtype.is_floating or x.placeholder.dtype.is_integer:
                        if f in (tf.reduce_mean, tf.reduce_sum):
                            replacement_value = zeros
                        elif f in (tf.reduce_max, tf.reduce_logsumexp, tf.argmax):
                            replacement_value = zeros + x.placeholder.dtype.min
                        elif f in (tf.reduce_min, tf.argmin):
                            replacement_value = zeros + x.placeholder.dtype.max
                        else:
                            raise ValueError("unexpected reduce function %r" % f)
                    elif x.placeholder.dtype.is_bool:
                        if f in (tf.reduce_any,):
                            replacement_value = zeros
                        elif f in (tf.reduce_all,):
                            replacement_value = tf.ones((), dtype=x.placeholder.dtype)
                        else:
                            raise ValueError("unexpected reduce function %r" % f)
                    else:
                        raise TypeError("reduce: unexpected input type %r from input %s" % (x.placeholder.dtype, x))

                    x_ = tf_util.where_bc(mask, x_, replacement_value, name="x_masked_axis_%i" % axis)
                    if f == tf.reduce_mean:
                        tag = x.dim_tags[axis]
                        assert tag.dyn_size_ext is not None  # checked above
                        size_all = tf.shape(x.placeholder)[axis]
                        size_actual = tag.dyn_size_ext
                        while any(d not in out_data.dim_tags for d in size_actual.dim_tags):
                            # We have some axis (e.g. B) which is not in the output.
                            # We need to remove this.
                            # https://github.com/rwth-i6/returnn/issues/1242
                            i, d = [(i, d) for i, d in enumerate(size_actual.dim_tags) if d not in out_data.dim_tags][0]
                            assert not d.is_dynamic()  # not implemented
                            size_all *= d.get_dim_value()
                            s = tf.reduce_sum(size_actual.placeholder, axis=i)
                            size_actual = size_actual.copy_template_excluding_axis(i)
                            size_actual.placeholder = s
                        seq_len_bc = size_actual.copy_compatible_to(
                            out_data, check_sparse=False, check_dtype=False
                        ).placeholder
                        seq_len_bc = tf.maximum(seq_len_bc, 1)  # avoid nan
                        correction_factor_ = tf.cast(size_all, tf.float32) / tf.cast(seq_len_bc, tf.float32)
                        correction_factor = tf_util.optional_mul(correction_factor, correction_factor_)
        if mode in arg_funcs:
            assert len(axes) == 1, "For argmax/argmin, only one reduction axis is supported"
            y = f(x_, axis=axes[0], output_type=tf.int32)
        else:
            y = f(x_, axis=axes)
            y = tf_util.optional_mul(y, correction_factor)
        out_data.raw_tensor = y
        return out_data



"""
Frontend API

The convention for the user is to do::

    from returnn import frontend as rf
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, TypeVar, Generic, Dict, Type, Union, Sequence
import numpy

from returnn.util.basic import NotSpecified
from returnn import tensor as _t

if TYPE_CHECKING:
    from returnn.tensor import Tensor, Dim
    from returnn._internal_frontend_api import InternalFrontend

RawTensorTypes = Union[int, float, complex, numpy.number, numpy.ndarray, bool, str]

T = TypeVar("T")  # tf.Tensor, torch.Tensor or so


class Frontend(Generic[T]):
    """
    Abstract base class for the frontend, operating on tensor type T, i.e. :class:`Tensor[T]`.

    This class and instances do not have any state,
    and all functions are staticmethod (or classmethod).
    """

    RawTensorType: Type[T]
    is_tensorflow: bool = False  # whether this framework uses TensorFlow
    _internal_frontend: Type[InternalFrontend[T]]

    @staticmethod
    def convert_to_tensor(value: Union[Tensor, T, RawTensorTypes]) -> Tensor[T]:
        """
        :param value: tensor, or scalar raw tensor or some other scalar value
        :return: tensor
        """
        raise NotImplementedError

    @classmethod
    def compare(
        cls,
        a: Union[Tensor, RawTensorTypes],
        kind: str,
        b: Union[Tensor, RawTensorTypes],
        *,
        allow_broadcast_all_sources: Optional[bool] = None,
        dim_order: Optional[Sequence[Dim]] = None,
    ) -> Tensor:
        """
        :param a:
        :param kind: "equal"|"==", "less"|"<", "less_equal"|"<=", "greater"|">", "greater_equal"|">=", "not_equal"|"!="
        :param b:
        :param allow_broadcast_all_sources: if True, it is allowed that neither a nor b has all dims of the result.
            Not needed when out_dims is specified explicitly.
        :param dim_order: defines the order of the resulting dims. if None, it is automatically inferred from a and b.
            Not all the dims of a and b need to be specified here, and there could also be other dims in the dim_order.
        :return: element-wise comparison of a and b
        """
        a = cls.convert_to_tensor(a)
        b = cls.convert_to_tensor(b)
        all_dims = []
        for dim in a.dims + b.dims:
            if dim not in all_dims:
                all_dims.append(dim)
        if all(set(x.dims) != set(all_dims) for x in (a, b)):
            if allow_broadcast_all_sources is False:
                raise ValueError(f"compare: sources {a!r} {b!r} not allowed with allow_broadcast_all_sources=False")
            elif allow_broadcast_all_sources is None:
                raise ValueError(f"compare: sources {a!r} {b!r} require explicit allow_broadcast_all_sources=True")
            elif allow_broadcast_all_sources is True:
                pass
            else:
                raise TypeError(f"invalid type for allow_broadcast_all_sources: {type(allow_broadcast_all_sources)}")
        if dim_order:
            all_dims.sort(key=lambda d: dim_order.index(d) if d in dim_order else len(dim_order))
        out = _t.Tensor("compare", dims=all_dims, dtype="bool")
        a = a.copy_compatible_to(out, check_sparse=False, check_dtype=False)
        b = b.copy_compatible_to(out, check_sparse=False, check_dtype=False)
        out.raw_tensor = cls._internal_frontend.compare_raw(a.raw_tensor, kind, b.raw_tensor)
        return out

    @staticmethod
    def dot(a: Tensor[T], b: Tensor[T], *, reduce: Union[Dim, Sequence[Dim]]) -> Tensor[T]:
        """
        This performs a dot-product of two sources a and b.
        The underlying operation is a batched matmul (shared..., I, J) * (shared..., J, K) -> (shared..., I, K).
        The inputs a and b are transformed internally into the required shapes in the following way:
        The axis J is specified via the Dim given as 'reduce'. If multiple reduce Dims are given the corresponding axes
        are merged into one before the matmul via a reshape. All other matching Dims in a and b will be treated as
        batch dimensions ('shared...'). Dims unique to a and b define the axes I and K, respectively. (Multiple or no
        unique axes in a and b are supported too.)

        Depending on which Dims exist in a, b and reduce this dot operation can be used to compute scaling, scalar
        product, outer product, matrix-vector multiplication, matrix-matrix multiplication etc. (all possibly batched).

        :param a:
        :param b:
        :param reduce: Dims over which to perform the product, have to be present in both a and b
        :return: result of dot product, Dim order: common axes as sorted in a, unique axes of a (in order),
            unique axes of b (in order)
        """
        raise NotImplementedError

    @staticmethod
    def range_over_dim(dim: Dim) -> Tensor[T]:
        """
        :param dim:
        :return: tensor with shape [dim]
        """
        raise NotImplementedError

    @staticmethod
    def reduce(
        source: Tensor[T],
        *,
        mode: str,
        axis: Union[Dim, Sequence[Dim]],
        use_time_mask: bool = NotSpecified,
    ) -> Tensor[T]:
        """
        Reduce the tensor along the given axis

        :param source:
        :param mode: "sum", "max", "min", "mean", "logsumexp", "any", "all", "argmin", "argmax"
        :param axis:
        :param use_time_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
        :return: tensor with axis removed
        """
        raise NotImplementedError


# We use a global instance, and we modify __class__ inplace,
# such that any reference to this can be updated.
# This is exposed to the user as `returnn.frontend`.
# The __class__ assignment is done in `select_engine`.
global_frontend = Frontend()

_dispatch_table = {}  # type: Dict[Type, Type[Frontend]]


def select_frontend_returnn_layers_tf():
    """
    Selects the RETURNN layers frontend (based on TF).
    """
    import tensorflow as tf

    frontend = get_frontend_by_tensor_type(tf.Tensor)  # side-effect: register it
    global_frontend.__class__ = frontend

    # TODO returnn layer type, register_frontend_by_tensor_type for that
    #   global_frontend.__class__ = ReturnnLayersFrontend


def select_frontend_torch():
    """
    Selects the PyTorch (low-level) frontend.
    """
    import torch

    frontend = get_frontend_by_tensor_type(torch.Tensor)  # side-effect: register it
    global_frontend.__class__ = frontend


def get_frontend_by_tensor_type(tensor_type: Type[T]) -> Type[Frontend[T]]:
    """
    :param tensor_type:
    """
    if tensor_type not in _dispatch_table:
        # We don't register all possible subclasses in the dispatch table.
        # Check through the MRO.
        for type_ in tensor_type.__mro__:
            if type_ in _dispatch_table:
                # Also register it for faster future lookups.
                _dispatch_table[tensor_type] = _dispatch_table[type_]
                return _dispatch_table[type_]
    if tensor_type not in _dispatch_table:
        # It would be registered if there was any select_engine or select_frontend_* call.
        # However, some code might not have done that, so for the common cases,
        # we do it here.
        if tensor_type.__module__.split(".")[0] == "tensorflow":
            from returnn.tf.frontend_low_level import TFFrontend

            frontend_type = TFFrontend
            tensor_types = _get_tensor_types_tf()
        elif tensor_type.__module__.split(".")[0] == "torch":
            from returnn.torch.frontend import TorchFrontend

            frontend_type = TorchFrontend
            tensor_types = _get_tensor_types_torch()
        else:
            raise Exception(f"unknown tensor type {tensor_type}")
        assert any(issubclass(tensor_type, type_) for type_ in tensor_types)
        for type_ in tensor_types:
            register_frontend_by_tensor_type(type_, frontend_type)
        return frontend_type
    return _dispatch_table[tensor_type]


def register_frontend_by_tensor_type(tensor_type: Type[T], frontend: Type[Frontend[T]]):
    """
    :param tensor_type:
    :param frontend:
    """
    _dispatch_table[tensor_type] = frontend


def _get_tensor_types_tf():
    """
    :return: tuple of relevant tensor types in TF.
        Note that it is not so important to cover all, as we also check issubclass as a fallback.
    """
    import tensorflow as tf

    ls = [tf.Tensor, tf.Variable]
    return tuple(ls)


def _get_tensor_types_torch():
    """
    :return: tuple of relevant tensor types in PyTorch.
        Note that it is not so important to cover all, as we also check issubclass as a fallback.
    """
    import torch

    ls = [torch.Tensor, torch.nn.Parameter]
    return tuple(ls)



"""
The main RETURNN package __init__.
We provide ``__version__`` and ``__long_version__``. See below.

You are supposed to explicitly import the specific sub-module/sub-package.
Just `import returnn` is not enough.

We also provide some helper code to keep older configs compatible,
which used our old-style module names, like ``import TFUtil`` or ``import returnn.TFUtil``.
"""

import os as _os

from .__setup__ import get_version_str as _get_version_str
from . import frontend_api as _frontend_api

__long_version__ = _get_version_str(
    fallback="1.0.0+unknown", long=True, verbose_error=True
)  # `SemVer <https://semver.org/>`__ compatible
__version__ = __long_version__[: __long_version__.index("+")]  # distutils.version.StrictVersion compatible
__git_version__ = __long_version__  # just an alias, to keep similar to other projects

from .__old_mod_loader__ import setup as _old_mod_loader_setup

_old_mod_loader_setup(modules=globals())

__root_dir__ = _os.path.dirname(_os.path.dirname(_os.path.abspath(__file__)))  # can be used as __path__

frontend = _frontend_api.global_frontend



#!/usr/bin/env python3

"""
Calculates word error rate (WER).
"""

from __future__ import annotations

import os
import sys
import time
import tensorflow as tf
import numpy
import typing

import _setup_returnn_env  # noqa
import returnn.__main__ as rnn
from returnn.log import log
import argparse
from returnn.util.basic import Stats, hms
from returnn.datasets.basic import Dataset, init_dataset
import returnn.util.basic as util
import returnn.tf.compat as tf_compat
import returnn.tf.util.basic as tf_util


class WerComputeGraph:
    """
    Creates TF computation graph to calculate the WER.
    We accumulate the absolute number of edits and normalize by the accumulated seq lens.
    """

    def __init__(self):
        self.hyps = tf_compat.v1.placeholder(tf.string, [None])
        self.refs = tf_compat.v1.placeholder(tf.string, [None])
        self.wer, self.ref_num_words = tf_util.string_words_calc_wer(hyps=self.hyps, refs=self.refs)
        self.total_wer_var = tf.Variable(initial_value=0, trainable=False, dtype=tf.int64)
        self.total_ref_num_words_var = tf.Variable(initial_value=0, trainable=False, dtype=tf.int64)
        self.update_total_wer = self.total_wer_var.assign_add(tf.reduce_sum(self.wer))
        self.update_ref_num_words = self.total_ref_num_words_var.assign_add(tf.reduce_sum(self.ref_num_words))
        self.updated_normalized_wer = tf.cast(self.update_total_wer, tf.float32) / tf.cast(
            self.update_ref_num_words, tf.float32
        )

    # noinspection PyShadowingNames
    def step(self, session, hyps, refs):
        """
        :param tf.compat.v1.Session session:
        :param list[str] hyps:
        :param list[str] refs:
        :return: updated normalized WER
        :rtype: float
        """
        return session.run(self.updated_normalized_wer, feed_dict={self.hyps: hyps, self.refs: refs})


def calc_wer_on_dataset(dataset, refs, options, hyps):
    """
    :param Dataset|None dataset:
    :param dict[str,str]|None refs: seq tag -> ref string (words delimited by space)
    :param options: argparse.Namespace
    :param dict[str,str] hyps: seq tag -> hyp string (words delimited by space)
    :return: WER
    :rtype: float
    """
    assert dataset or refs
    start_time = time.time()
    seq_len_stats = {"refs": Stats(), "hyps": Stats()}
    seq_idx = options.startseq
    if options.endseq < 0:
        options.endseq = float("inf")
    wer = 1.0
    remaining_hyp_seq_tags = set(hyps.keys())
    interactive = util.is_tty() and not log.verbose[5]
    collected = {"hyps": [], "refs": []}
    max_num_collected = 1
    if dataset:
        dataset.init_seq_order(epoch=1)
    else:
        refs = sorted(refs.items(), key=lambda item: len(item[1]))
    while True:
        if seq_idx > options.endseq:
            break
        if dataset:
            if not dataset.is_less_than_num_seqs(seq_idx):
                break
            dataset.load_seqs(seq_idx, seq_idx + 1)
            complete_frac = dataset.get_complete_frac(seq_idx)
            seq_tag = dataset.get_tag(seq_idx)
            assert isinstance(seq_tag, str)
            ref = dataset.get_data(seq_idx, options.key)
            if isinstance(ref, numpy.ndarray):
                assert ref.shape == ()
                ref = ref.flatten()[0]  # get the entry itself (str or bytes)
            if isinstance(ref, bytes):
                ref = ref.decode("utf8")
            assert isinstance(ref, str)
            try:
                num_seqs_s = str(dataset.num_seqs)
            except NotImplementedError:
                try:
                    num_seqs_s = "~%i" % dataset.estimated_num_seqs
                except TypeError:  # a number is required, not NoneType
                    num_seqs_s = "?"
        else:
            if seq_idx >= len(refs):
                break
            complete_frac = (seq_idx + 1) / float(len(refs))
            seq_tag, ref = refs[seq_idx]
            assert isinstance(seq_tag, str)
            assert isinstance(ref, str)
            num_seqs_s = str(len(refs))

        start_elapsed = time.time() - start_time
        progress_prefix = "%i/%s (WER %.02f%%)" % (seq_idx, num_seqs_s, wer * 100)
        progress = "%s (%.02f%%)" % (progress_prefix, complete_frac * 100)
        if complete_frac > 0:
            total_time_estimated = start_elapsed / complete_frac
            remaining_estimated = total_time_estimated - start_elapsed
            progress += " (%s)" % hms(remaining_estimated)

        remaining_hyp_seq_tags.remove(seq_tag)
        hyp = hyps[seq_tag]
        seq_len_stats["hyps"].collect([len(hyp)])
        seq_len_stats["refs"].collect([len(ref)])
        collected["hyps"].append(hyp)
        collected["refs"].append(ref)

        if len(collected["hyps"]) >= max_num_collected:
            wer = wer_compute.step(session, **collected)
            del collected["hyps"][:]
            del collected["refs"][:]

        if interactive:
            util.progress_bar_with_time(complete_frac, prefix=progress_prefix)
        elif log.verbose[5]:
            print(progress_prefix, "seq tag %r, ref/hyp len %i/%i chars" % (seq_tag, len(ref), len(hyp)))
        seq_idx += 1
    if len(collected["hyps"]) > 0:
        wer = wer_compute.step(session, **collected)
    print("Done. Num seqs %i. Total time %s." % (seq_idx, hms(time.time() - start_time)), file=log.v1)
    print("Remaining num hyp seqs %i." % (len(remaining_hyp_seq_tags),), file=log.v1)
    if dataset:
        print("More seqs which we did not dumped: %s." % dataset.is_less_than_num_seqs(seq_idx), file=log.v1)
    for key in ["hyps", "refs"]:
        seq_len_stats[key].dump(stream_prefix="Seq-length %r %r " % (key, options.key), stream=log.v2)
    if options.expect_full:
        assert not remaining_hyp_seq_tags, "There are still remaining hypotheses."
    return wer


config = None  # type: typing.Optional["returnn.config.Config"]


def init(config_filename, log_verbosity):
    """
    :param str config_filename: filename to config-file
    :param int log_verbosity:
    """
    rnn.init_better_exchook()
    rnn.init_thread_join_hack()
    if config_filename:
        print("Using config file %r." % config_filename)
        assert os.path.exists(config_filename)
    rnn.init_config(config_filename=config_filename, command_line_options=[])
    global config
    config = rnn.config
    config.set("task", "calculate_wer")
    config.set("log", None)
    config.set("log_verbosity", log_verbosity)
    config.set("use_tensorflow", True)
    rnn.init_log()
    print("Returnn calculate-word-error-rate starting up.", file=log.v1)
    rnn.returnn_greeting()
    rnn.init_backend_engine()
    assert util.BackendEngine.is_tensorflow_selected(), "this is only for TensorFlow"
    rnn.init_faulthandler()
    rnn.init_config_json_network()
    rnn.print_task_properties()


def load_hyps_refs(filename):
    """
    :param str filename:
    :return: dict of seq_tag -> ref
    :rtype: dict[str,str]
    """
    if filename.endswith(".gz"):
        import gzip

        content_str = gzip.open(filename, "rt").read()
    else:
        content_str = open(filename).read()
    content = eval(content_str)
    # See dump-dataset-raw-strings.py.
    # We expect that it is stored as a dict.
    assert isinstance(content, dict)
    assert len(content) > 0
    example_hyp = next(iter(content.items()))
    assert isinstance(example_hyp[0], str)  # seq tag
    if isinstance(example_hyp[1], list):
        assert isinstance(example_hyp[1][0][1], str)
        # n-best list output format needs to be converted first
        # always pick the best (first) entry from the list (which is [(score, text), ...]
        content = {seq_tag: nbest_list[0][1] for seq_tag, nbest_list in content.items()}
    else:
        assert isinstance(example_hyp[1], str)
    return content


wer_compute = None  # type: typing.Optional[WerComputeGraph]
session = None  # type: typing.Optional[tf.compat.v1.Session]


def main(argv):
    """
    Main entry.
    """
    arg_parser = argparse.ArgumentParser(description="Dump something from dataset.")
    arg_parser.add_argument("--config", help="filename to config-file. will use dataset 'eval' from it")
    arg_parser.add_argument("--dataset", help="dataset, overwriting config")
    arg_parser.add_argument("--refs", help="same format as hyps. alternative to providing dataset/config")
    arg_parser.add_argument("--hyps", help="hypotheses, dumped via search in py format")
    arg_parser.add_argument("--startseq", type=int, default=0, help="start seq idx (inclusive) (default: 0)")
    arg_parser.add_argument("--endseq", type=int, default=-1, help="end seq idx (inclusive) or -1 (default: -1)")
    arg_parser.add_argument("--key", default="raw", help="data-key, e.g. 'data' or 'classes'. (default: 'raw')")
    arg_parser.add_argument("--verbosity", default=4, type=int, help="5 for all seqs (default: 4)")
    arg_parser.add_argument("--out", help="if provided, will write WER% (as string) to this file")
    arg_parser.add_argument("--expect_full", action="store_true", help="full dataset should be scored")
    args = arg_parser.parse_args(argv[1:])
    assert args.config or args.dataset or args.refs

    init(config_filename=args.config, log_verbosity=args.verbosity)
    dataset = None
    refs = None
    if args.refs:
        refs = load_hyps_refs(args.refs)
    elif args.dataset:
        dataset = init_dataset(args.dataset)
    elif config.value("wer_data", "eval") in ["train", "dev", "eval"]:
        dataset = init_dataset(config.opt_typed_value(config.value("search_data", "eval")))
    else:
        dataset = init_dataset(config.opt_typed_value("wer_data"))
    hyps = load_hyps_refs(args.hyps)

    global wer_compute
    wer_compute = WerComputeGraph()
    with tf_compat.v1.Session(config=tf_compat.v1.ConfigProto(device_count={"GPU": 0})) as _session:
        global session
        session = _session
        session.run(tf_compat.v1.global_variables_initializer())
        try:
            wer = calc_wer_on_dataset(dataset=dataset, refs=refs, options=args, hyps=hyps)
            print("Final WER: %.02f%%" % (wer * 100), file=log.v1)
            if args.out:
                with open(args.out, "w") as output_file:
                    output_file.write("%.02f\n" % (wer * 100))
                print("Wrote WER%% to %r." % args.out)
        except KeyboardInterrupt:
            print("KeyboardInterrupt")
            sys.exit(1)
        finally:
            rnn.finalize()


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python3

"""
Will use the PyCharm code inspection.

See here:
  https://github.com/albertz/pycharm-inspect
  https://stackoverflow.com/questions/55323910/pycharm-code-style-check-via-command-line
  https://youtrack.jetbrains.com/issue/PY-34863
  https://youtrack.jetbrains.com/issue/PY-34864
"""

import os
import sys
import re
import time
import shutil
import subprocess
import tempfile
import typing
from glob import glob
import argparse
from xml.dom import minidom
from xml.etree import ElementTree

my_dir = os.path.dirname(os.path.abspath(__file__))
root_dir = os.path.dirname(my_dir)
sys.path.insert(0, root_dir)
os.chdir(root_dir)

from returnn.util import better_exchook  # noqa
from returnn.util.basic import pip_install, which_pip, pip_check_is_installed, hms  # noqa

travis_env = os.environ.get("TRAVIS") == "true"
github_env = os.environ.get("GITHUB_ACTIONS") == "true"

gray_color = "black"  # black is usually gray
if github_env:
    gray_color = "white"  # black is black, on black background. so just use white


class _StdoutTextFold:
    def __init__(self, name):
        """
        :param str name:
        """
        self.name = name
        self.start_time = time.time()

        if github_env:
            # https://github.community/t/has-github-action-somthing-like-travis-fold/16841
            if not folds:  # nested folds not supported, https://github.com/actions/toolkit/issues/112
                print("::group::%s" % name)

        if travis_env:
            # travis_fold: https://github.com/travis-ci/travis-ci/issues/1065
            print("travis_fold:start:%s" % name)

        sys.stdout.flush()

    def finish(self):
        """
        End fold.
        """
        elapsed_time = time.time() - self.start_time
        print("%s: Elapsed time: %s" % (self.name, hms(elapsed_time)))

        if travis_env:
            print("travis_fold:end:%s" % folds[-1])

        if github_env:
            if len(folds) == 1:
                print("::endgroup::")

        sys.stdout.flush()


folds = []  # type: typing.List[_StdoutTextFold]


def fold_start(name):
    """
    :param str name:
    """
    folds.append(_StdoutTextFold(name))


def fold_end():
    """
    Ends the fold.
    """
    assert folds
    folds[-1].finish()
    folds.pop(-1)


def check_pycharm_dir(pycharm_dir):
    """
    :param str pycharm_dir:
    """
    assert os.path.isdir(pycharm_dir)
    assert os.path.exists("%s/bin/inspect.sh" % pycharm_dir)


def install_pycharm():
    """
    :return: pycharm dir
    :rtype: str
    """
    fold_start("script.install")
    print("travis_fold:start:script.install")
    install_dir = tempfile.mkdtemp()
    pycharm_dir = "%s/pycharm" % install_dir
    print("Install PyCharm into:", pycharm_dir)
    sys.stdout.flush()

    pycharm_version = (2020, 2)
    name = "pycharm-community-%i.%i" % pycharm_version
    fn = "%s.tar.gz" % name

    subprocess.check_call(
        ["wget", "--progress=dot:mega", "-c", "https://download.jetbrains.com/python/%s" % fn],
        cwd=install_dir,
        stderr=subprocess.STDOUT,
    )
    tar_out = subprocess.check_output(["tar", "-xzvf", fn], cwd=install_dir, stderr=subprocess.STDOUT)
    print((b"\n".join(tar_out.splitlines()[-10:])).decode("utf8"))
    assert os.path.isdir("%s/%s" % (install_dir, name))
    os.remove("%s/%s" % (install_dir, fn))
    os.rename("%s/%s" % (install_dir, name), pycharm_dir)
    check_pycharm_dir(pycharm_dir)

    fold_end()
    return pycharm_dir


def get_version_str_from_pycharm(pycharm_dir):
    """
    :param str pycharm_dir:
    :return: e.g. "CE2018.3"
    :rtype: str
    """
    import re
    import json

    if os.path.exists("%s/product-info.json" % pycharm_dir):
        d = json.load(open("%s/product-info.json" % pycharm_dir))
        name = d["dataDirectoryName"]
        assert isinstance(name, str)
        assert name.startswith("PyCharm")
        return name[len("PyCharm") :]
    # This works on PyCharm 2019.
    code = open("%s/bin/pycharm.sh" % pycharm_dir).read()
    m = re.search("-Didea\\.paths\\.selector=PyCharm(\\S+) ", code)
    assert m, "pycharm %r not as expected" % pycharm_dir
    return m.group(1)


def parse_pycharm_version(version_str):
    """
    :param str version_str: e.g. "CE2018.3"
    :rtype: ((int,int),str)
    :return: e.g. (2018,3),"CE"
    """
    name = ""
    if version_str.startswith("CE"):
        name = "CE"
        version_str = version_str[2:]
    assert version_str.startswith("2")
    version_str_parts = version_str.split(".")
    assert len(version_str_parts) == 2, "version %r" % version_str
    return tuple([int(p) for p in version_str_parts]), name


def create_stub_dir(pycharm_dir, stub_dir, pycharm_major_version):
    """
    :param str pycharm_dir:
    :param str stub_dir:
    :param int pycharm_major_version:
    """
    fold_start("script.create_python_stubs")
    print("Generating Python stubs via helpers/generator3.py...")
    if pycharm_major_version >= 2020:
        generator_path = "%s/plugins/python-ce/helpers/generator3/__main__.py" % pycharm_dir
        assert os.path.exists(generator_path)
        cmd = [sys.executable, generator_path, "-d", stub_dir]
        # The stdout can sometimes be very long. Thus we pipe and filter it a bit.
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        stdout, _ = proc.communicate()
        if proc.returncode != 0:
            raise subprocess.CalledProcessError(returncode=proc.returncode, cmd=cmd, output=stdout)
        for line in stdout.splitlines():
            line = line.decode("utf8")
            if len(line) < 240:
                print(line)
            else:
                print(line[:240] + "...")
    elif pycharm_major_version <= 2019:
        generator_path = "%s/helpers/generator3.py" % pycharm_dir
        assert os.path.exists(generator_path)
        subprocess.check_call([sys.executable, generator_path, "-d", stub_dir, "-b"])
        print("Collecting further native modules...")
        sys.stdout.flush()
        mod_names = []
        for line in subprocess.check_output([sys.executable, generator_path, "-L"]).decode("utf8").splitlines()[1:]:
            # First line is version, so we skipped those.
            # Then we get sth like "<module name> <other things>...".
            assert isinstance(line, str)
            mod_name = line.split()[0]
            # There are duplicates. Ignore.
            if mod_name not in mod_names:
                mod_names.append(mod_name)
        for mod_name in mod_names:
            print("Generate for %r." % mod_name)
            sys.stdout.flush()
            # Ignore errors here.
            subprocess.call([sys.executable, generator_path, "-d", stub_dir, mod_name])
    fold_end()


_use_stub_zip = False


def setup_pycharm_python_interpreter(pycharm_dir):
    """
    Unfortunately, the headless PyCharm bin/inspect will use the global PyCharm settings,
    and requires that we have a Python interpreter set up,
    with the same name as we use in our `.idea` settings, which we will link in :func:`prepare_src_dir`.
    See here: https://youtrack.jetbrains.com/issue/PY-34864

    Our current way to work around this: We create (or extend) the file
    ``~/.PyCharm<VERSION>/config/options/jdk.table.xml`` such that it has the right Python interpreter.

    :param str pycharm_dir:
    """
    fold_start("script.setup_pycharm_python_interpreter")
    print("Setup PyCharm Python interpreter... (jdk.table.xml)")
    print("Current Python:", sys.executable, sys.version, sys.version_info)
    name = "Python 3 (.../bin/python3)"  # used in our PyCharm.idea. this should match.
    pycharm_version_str = get_version_str_from_pycharm(pycharm_dir)
    pycharm_version, pycharm_version_name = parse_pycharm_version(pycharm_version_str)
    if sys.platform == "darwin":
        pycharm_config_dir = os.path.expanduser("~/Library/Preferences/PyCharm%s" % pycharm_version_str)
        pycharm_system_dir = os.path.expanduser("~/Library/Caches/PyCharm%s" % pycharm_version_str)
    else:  # assume Linux/Unix
        if pycharm_version[0] >= 2020:
            pycharm_config_dir = os.path.expanduser("~/.config/JetBrains/PyCharm%s" % pycharm_version_str)
            pycharm_system_dir = os.path.expanduser("~/.cache/JetBrains/PyCharm%s" % pycharm_version_str)
        else:  # <= 2020
            pycharm_config_dir = os.path.expanduser("~/.PyCharm%s/config" % pycharm_version_str)
            pycharm_system_dir = os.path.expanduser("~/.PyCharm%s/system" % pycharm_version_str)

    # I just zipped the stubs from my current installation on Linux.
    # Maybe we can also reuse these stubs for other PyCharm versions, or even other Python versions.
    if _use_stub_zip:
        stub_base_name = "pycharm2018.3-python3.6-stubs"
        stub_fn = "%s/python_stubs/%s.zip" % (pycharm_system_dir, stub_base_name)
        stub_dir = "%s/python_stubs/%s" % (pycharm_system_dir, stub_base_name)
        os.makedirs(os.path.dirname(stub_fn), exist_ok=True)
        if os.path.exists(stub_dir):
            print("Python stubs dir exists already:", stub_dir)
        else:
            if not os.path.exists(stub_fn):
                subprocess.check_call(
                    ["wget", "https://www-i6.informatik.rwth-aachen.de/web/Software/returnn/%s.zip" % stub_base_name],
                    cwd=os.path.dirname(stub_fn),
                )
            assert os.path.exists(stub_fn)
            subprocess.check_call(
                ["unzip", "%s.zip" % stub_base_name, "-d", stub_base_name], cwd=os.path.dirname(stub_fn)
            )
            assert os.path.isdir(stub_dir)
    else:
        fold_start("script.opt_install_further_py_deps")
        if not pip_check_is_installed("tensorflow") and not pip_check_is_installed("tensorflow-gpu"):
            pip_install("tensorflow")
        # Note: Horovod will usually fail to install in this env.
        for pkg in ["typing", "librosa==0.8.1", "PySoundFile", "nltk", "matplotlib", "mpi4py", "pycodestyle"]:
            if not pip_check_is_installed(pkg):
                try:
                    pip_install(pkg)
                except subprocess.CalledProcessError as exc:
                    print("Pip install failed:", exc)
                    print("Ignore...")
        fold_end()

        stub_dir = "%s/python_stubs/python%s-generated" % (pycharm_system_dir, "%i.%i.%i" % sys.version_info[:3])
        if os.path.exists(stub_dir):
            print("Python stubs already exists, not recreating (%s)" % stub_dir)
        else:
            print("Generate stub dir:", stub_dir)
            os.makedirs(stub_dir)
            create_stub_dir(pycharm_dir=pycharm_dir, stub_dir=stub_dir, pycharm_major_version=pycharm_version[0])

    jdk_table_fn = "%s/options/jdk.table.xml" % pycharm_config_dir
    print("Filename:", jdk_table_fn)
    os.makedirs(os.path.dirname(jdk_table_fn), exist_ok=True)

    if os.path.exists(jdk_table_fn):
        print("Loading existing jdk.table.xml.")
        et = ElementTree.parse(jdk_table_fn)
        root = et.getroot()
        assert isinstance(root, ElementTree.Element)
        jdk_collection = root.find("./component")
        assert isinstance(jdk_collection, ElementTree.Element)
        assert jdk_collection.tag == "component" and jdk_collection.attrib["name"] == "ProjectJdkTable"
    else:
        print("Creating new jdk.table.xml.")
        root = ElementTree.Element("application")
        et = ElementTree.ElementTree(root)
        jdk_collection = ElementTree.SubElement(root, "component", name="ProjectJdkTable")
        assert isinstance(jdk_collection, ElementTree.Element)

    existing_jdk = jdk_collection.find("./jdk/name[@value='%s']/.." % name)
    if existing_jdk:
        print("Found existing Python interpreter %r. Remove and recreate." % name)
        assert isinstance(existing_jdk, ElementTree.Element)
        assert existing_jdk.find("./name").attrib["value"] == name
        jdk_collection.remove(existing_jdk)

    # Example content:
    """
  <application>
  <component name="ProjectJdkTable">
    <jdk version="2">
      <name value="Python 2.7.3 (/usr/bin/python2.7)" />
      <type value="Python SDK" />
      <version value="Python 2.7.12" />
      <homePath value="/usr/bin/python2.7" />
      <roots>
        <classPath>
          <root type="composite">
            <root url="file:///usr/bin" type="simple" />
            ...
          </root>
        </classPath>
        <sourcePath>
          <root type="composite" />
        </sourcePath>
      </roots>
      <additional />
    </jdk>
  </component>
  </application>
  """

    jdk_entry = ElementTree.SubElement(jdk_collection, "jdk", version="2")
    ElementTree.SubElement(jdk_entry, "name", value=name)
    ElementTree.SubElement(jdk_entry, "type", value="Python SDK")
    ElementTree.SubElement(jdk_entry, "version", value="Python %i.%i.%i" % sys.version_info[:3])
    ElementTree.SubElement(jdk_entry, "homePath", value=sys.executable)
    paths_root = ElementTree.SubElement(jdk_entry, "roots")
    classes_paths = ElementTree.SubElement(ElementTree.SubElement(paths_root, "classPath"), "root", type="composite")
    relevant_paths = list(sys.path)
    if root_dir in relevant_paths:
        relevant_paths.remove(root_dir)
    if my_dir in relevant_paths:
        relevant_paths.remove(my_dir)
    relevant_paths.extend(
        [
            stub_dir,
            "$APPLICATION_HOME_DIR$/helpers/python-skeletons",
            "$APPLICATION_HOME_DIR$/helpers/typeshed/stdlib/3",
            "$APPLICATION_HOME_DIR$/helpers/typeshed/stdlib/2and3",
            "$APPLICATION_HOME_DIR$/helpers/typeshed/third_party/3",
            "$APPLICATION_HOME_DIR$/helpers/typeshed/third_party/2and3",
        ]
    )
    # Maybe also add Python stubs path? How to generate them?
    for path in relevant_paths:
        ElementTree.SubElement(classes_paths, "root", url="file://%s" % path, type="simple")
    ElementTree.SubElement(ElementTree.SubElement(paths_root, "sourcePath"), "root", type="composite")
    ElementTree.SubElement(jdk_entry, "additional")

    print("Save XML.")
    et.write(jdk_table_fn, encoding="UTF-8")

    fold_start("script.jdk_table")
    print("XML content:")
    rough_string = ElementTree.tostring(root, "utf-8")
    print(minidom.parseString(rough_string).toprettyxml(indent="  "))
    fold_end()

    fold_end()


def read_spelling_dict():
    """
    :rtype: list[str]
    """
    return open("%s/spelling.dic" % my_dir).read().splitlines()


def create_spelling_dict_xml(src_dir):
    """
    Need to create this on-the-fly for the current user.
    """
    # Example:
    """
  <component name="ProjectDictionaryState">
  <dictionary name="az">
    <words>
      <w>dtype</w>
      <w>idxs</w>
      <w>keepdims</w>
      ...
    </words>
  </dictionary>
  </component>
  """
    from returnn.util.basic import get_login_username

    user_name = get_login_username()
    root = ElementTree.Element("component", name="ProjectDictionaryState")
    dict_ = ElementTree.SubElement(root, "dictionary", name=user_name)
    words = ElementTree.SubElement(dict_, "words")
    for w in read_spelling_dict():
        ElementTree.SubElement(words, "w").text = w
    et = ElementTree.ElementTree(root)
    print("Save XML.")
    xml_filename = "%s/.idea/dictionaries/%s.xml" % (src_dir, user_name)
    os.makedirs(os.path.dirname(xml_filename), exist_ok=True)
    et.write(xml_filename, encoding="UTF-8")


def prepare_src_dir(files=None):
    """
    New clean source dir, where we symlink only the relevant src files.

    :param list[str]|None files:
    :return: src dir
    :rtype: str
    """
    fold_start("script.prepare")
    print("Prepare project source files...")
    if not files:
        files = ["returnn", "tools", "demos", "rnn.py", "setup.py", "__init__.py"]
    src_tmp_dir = "%s/returnn" % tempfile.mkdtemp()
    os.mkdir(src_tmp_dir)
    shutil.copytree("%s/PyCharm.idea" % my_dir, "%s/.idea" % src_tmp_dir, symlinks=True)
    for fn in files:
        fn = "%s/%s" % (root_dir, fn)
        dst = "%s/%s" % (src_tmp_dir, os.path.basename(fn))
        if os.path.isdir(fn):
            shutil.copytree(fn, dst, symlinks=True)
        else:
            shutil.copy(fn, dst)
    create_spelling_dict_xml(src_tmp_dir)
    print("All source files:")
    sys.stdout.flush()
    subprocess.check_call(["ls", "-la", src_tmp_dir])
    fold_end()
    return src_tmp_dir


def run_inspect(pycharm_dir, src_dir, skip_pycharm_inspect=False):
    """
    :param str pycharm_dir:
    :param str src_dir:
    :param bool skip_pycharm_inspect:
    :return: dir of xml files
    :rtype: str
    """
    out_tmp_dir = tempfile.mkdtemp()

    fold_start("script.inspect")
    if not skip_pycharm_inspect:
        # Note: Will not run if PyCharm is already running.
        # Maybe we can find some workaround for this?
        # See here: https://stackoverflow.com/questions/55339010/run-pycharm-inspect-sh-even-if-pycharm-is-already-running
        # And here: https://github.com/albertz/pycharm-inspect
        # Also: https://stackoverflow.com/questions/55323910/pycharm-code-style-check-via-command-line
        cmd = [
            "%s/bin/inspect.sh" % pycharm_dir,
            src_dir,
            "%s/PyCharm-inspection-profile.xml" % my_dir,
            out_tmp_dir,
            "-v2",
        ]
        print("$ %s" % " ".join(cmd))
        subprocess.check_call(cmd, stderr=subprocess.STDOUT)

    # PyCharm does not do PEP8 code style checks by itself but uses the (bundled) pycodestyle tool.
    # https://youtrack.jetbrains.com/issue/PY-43901
    # Do that now.
    root = ElementTree.Element("problems")
    from lint_common import find_all_py_source_files

    for py_src_file in find_all_py_source_files():
        ignore_codes = "E121,E123,E126,E226,E24,E704,W503,W504"  # PyCharm defaults
        ignore_codes += (
            ",E203"  # https://black.readthedocs.io/en/stable/faq.html#why-are-flake8-s-e203-and-w503-violated
        )
        indent_size = 4
        cmd = [
            "pycodestyle",
            py_src_file,
            "--ignore=%s" % ignore_codes,
            "--indent-size=%i" % indent_size,
            "--max-line-length=120",
        ]
        print("$ %s" % " ".join(cmd))
        sys.stdout.flush()
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        stdout, _ = proc.communicate()
        # We do not check returncode, as this is always non-zero if there is any warning.
        for line in stdout.decode("utf8").splitlines():
            # Example line: demos/demo-record-and-push-to-webserver.py:48:1: E302 expected 2 blank lines, found 1
            m = re.match("^(.*):([0-9]+):([0-9]+): ([EW][0-9]+) (.+)$", line)
            assert m, "unexpected line %r" % line
            fn_, line_nr, col_nr, warn_id, description = m.groups()
            assert fn_ == py_src_file, "unexpected line %r" % line
            line_nr, col_nr = int(line_nr), int(col_nr)
            description = "%s: %s" % (warn_id, description)
            prob = ElementTree.SubElement(root, "problem")
            # Note: We do not aim to have this complete. This is just enough such that report_inspect_xml can read it.
            ElementTree.SubElement(prob, "file").text = "file://$PROJECT_DIR$/%s" % py_src_file
            ElementTree.SubElement(prob, "line").text = str(line_nr)
            ElementTree.SubElement(prob, "offset").text = str(col_nr)
            ElementTree.SubElement(prob, "problem_class", severity="WEAK WARNING", id=warn_id).text = description
            ElementTree.SubElement(prob, "description").text = description
    et = ElementTree.ElementTree(root)
    et.write("%s/Pep8CodeStyle.xml" % out_tmp_dir, encoding="UTF-8")

    fold_end()
    return out_tmp_dir


def report_inspect_xml(fn):
    """
    :param str fn:
    :return: list of (filename, line, problem_severity, inspect_class, description)
    :rtype: list[(str,int,str,str,str)]
    """
    # Example PyPackageRequirementsInspection.xml:
    """
  <problems is_local_tool="true">
  <problem>
    <file>file://$PROJECT_DIR$/TFUtil.py</file>
    <line>1</line>
    <module>returnn</module>
    <entry_point TYPE="file" FQNAME="file://$PROJECT_DIR$/TFUtil.py" />
    <problem_class severity="WARNING" attribute_key="WARNING_ATTRIBUTES">Package requirements</problem_class>
    <description>Package requirements 'h5py', 'theano==0.9' are not satisfied</description>
  </problem>
  </problems>
  """
    inspect_class = os.path.splitext(os.path.basename(fn))[0]  # e.g. "PyPackageRequirementsInspection"
    root = ElementTree.parse(fn).getroot()
    assert isinstance(root, ElementTree.Element)
    assert root.tag == "problems"
    result = []
    for problem in root.findall("./problem"):
        assert isinstance(problem, ElementTree.Element)
        assert problem.tag == "problem"
        filename = problem.find("./file").text.strip()
        if filename.startswith("file://$PROJECT_DIR$/"):
            filename = filename[len("file://$PROJECT_DIR$/") :]
        line = int(problem.find("./line").text.strip())
        problem_severity = problem.find("./problem_class").attrib["severity"]
        description = problem.find("./description").text.strip()

        # Do some filtering for false positives. This is ugly, but the other solution would be to ignore all of them.
        possible_false_positive = False
        if inspect_class == "PyArgumentListInspection" and "'d0' unfilled" in description:  # Numpy false positive
            possible_false_positive = True
        if inspect_class == "PyArgumentListInspection" and "'d1' unfilled" in description:  # Numpy false positive
            possible_false_positive = True
        if inspect_class == "PyArgumentListInspection" and "'self' unfilled" in description:  # Numpy false positive
            possible_false_positive = True
        if inspect_class == "PyStringFormatInspection" and "Unexpected type None" in description:
            possible_false_positive = True
        if possible_false_positive:
            problem_severity = "POSSIBLE-FALSE %s" % problem_severity

        result.append((filename, line, problem_severity, inspect_class, description))

    return result


def report_inspect_dir(
    inspect_xml_dir, inspect_class_blacklist=None, inspect_class_not_counted=None, ignore_count_for_files=()
):
    """
    :param str inspect_xml_dir:
    :param set[str]|None inspect_class_blacklist:
    :param set[str]|None inspect_class_not_counted:
    :param set[str]|tuple[str]|None ignore_count_for_files:
    :return: count of reports
    :rtype: int
    """
    if os.path.isfile(inspect_xml_dir):
        assert inspect_xml_dir.endswith(".xml")
        inspect_xml_files = [inspect_xml_dir]
    else:
        assert os.path.isdir(inspect_xml_dir)
        inspect_xml_files = list(glob(inspect_xml_dir + "/*.xml"))
        assert inspect_xml_files

    inspections = []
    for fn in inspect_xml_files:
        inspections.extend(report_inspect_xml(fn))
    inspections.sort()
    inspections.append((None, None, None, None, None))  # final marker

    # copy
    inspect_class_blacklist = set(inspect_class_blacklist or ())
    inspect_class_not_counted = set(inspect_class_not_counted or ())

    # maybe update inspect_class_not_counted
    from lint_common import find_all_py_source_files

    returnn_py_source_files = set(find_all_py_source_files())
    all_files = set()
    relevant_inspections_for_file = set()
    explicitly_ignored_files = ignore_count_for_files
    ignore_count_for_files = set(ignore_count_for_files)
    for filename, line, problem_severity, inspect_class, description in inspections:
        all_files.add(filename)
        if filename not in returnn_py_source_files:
            continue
        if inspect_class in inspect_class_blacklist:
            continue
        if inspect_class in inspect_class_not_counted:
            continue
        if problem_severity.startswith("POSSIBLE-FALSE "):
            continue
        relevant_inspections_for_file.add(filename)
    for filename in all_files:
        if filename not in relevant_inspections_for_file:
            ignore_count_for_files.add(filename)

    print("Reporting individual files. We skip all files which have no warnings at all.")
    color = better_exchook.Color()
    total_relevant_count = 0
    file_count = None
    last_filename = None
    for filename, line, problem_severity, inspect_class, description in inspections:
        if filename and filename not in returnn_py_source_files:
            continue  # for now, to not spam Travis too much
        if inspect_class in inspect_class_blacklist:
            continue

        if filename != last_filename:
            if last_filename:
                if last_filename in explicitly_ignored_files:
                    msg = color.color("This file is on the ignore list.", color=gray_color)
                elif last_filename not in returnn_py_source_files:
                    msg = color.color(
                        "This file is not part of the official RETURNN Python source code.", color=gray_color
                    )
                elif last_filename in ignore_count_for_files:
                    msg = color.color("The inspection reports for this file are all non critical.", color=gray_color)
                else:
                    msg = color.color("The inspection reports for this file are fatal!", color="red")
                print(msg)
                fold_end()
            if filename:
                file_msg = color.color(
                    "File: %s" % filename, color=gray_color if filename in ignore_count_for_files else "red"
                )
                if github_env:
                    fold_start(file_msg)
                else:
                    fold_start("inspect.%s" % filename)
                    print(file_msg)
            last_filename = filename
            file_count = 0
        if not filename:
            continue
        if filename in ignore_count_for_files and file_count >= 10:
            if file_count == 10:
                print("... (further warnings skipped)")
            file_count += 1
            continue

        msg = "%s:%i: %s %s: %s" % (filename, line, problem_severity, inspect_class, description)
        msg_counted = True
        if inspect_class in inspect_class_not_counted:
            msg_counted = False
        if problem_severity.startswith("POSSIBLE-FALSE "):
            msg_counted = False
        if msg_counted:
            print(color.color(msg, color="red"))
            if filename not in ignore_count_for_files:
                total_relevant_count += 1
        else:
            print(color.color(msg, color=gray_color))
        file_count += 1

    print("Total relevant inspection reports:", total_relevant_count)
    return total_relevant_count


def main():
    """
    Main entry point for this script.
    """
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("--xml")
    arg_parser.add_argument("--pycharm")
    arg_parser.add_argument("--setup_pycharm_only", action="store_true")
    arg_parser.add_argument("--skip_setup_pycharm", action="store_true")
    arg_parser.add_argument("--skip_pycharm_inspect", action="store_true", help="only PEP8")
    arg_parser.add_argument("--files", nargs="*")
    args = arg_parser.parse_args()

    from lint_common import ignore_count_for_files

    inspect_kwargs = dict(
        inspect_class_blacklist={},
        inspect_class_not_counted={
            # Here we disable more than what you would do in the IDE.
            # The aim is that any left over warnings are always indeed important and should be fixed.
            # False alarms.
            "PyTypeCheckerInspection",  # too much false alarms: https://youtrack.jetbrains.com/issue/PY-34893
            # Not critical.
            "SpellCheckingInspection",  # way too much for now...
            "PyClassHasNoInitInspection",  # not relevant?
            "PyMethodMayBeStaticInspection",  # not critical
        },
        ignore_count_for_files=ignore_count_for_files,
    )

    if args.xml:
        if report_inspect_dir(args.xml, **inspect_kwargs) > 0:
            sys.exit(1)
        return

    if args.pycharm:
        pycharm_dir = args.pycharm
        check_pycharm_dir(pycharm_dir)
    else:
        pycharm_dir = install_pycharm()

    if not args.skip_setup_pycharm and not args.skip_pycharm_inspect:
        setup_pycharm_python_interpreter(pycharm_dir=pycharm_dir)
    if args.setup_pycharm_only:
        return

    src_dir = prepare_src_dir(files=args.files)
    res_dir = run_inspect(pycharm_dir=pycharm_dir, src_dir=src_dir, skip_pycharm_inspect=args.skip_pycharm_inspect)
    if report_inspect_dir(res_dir, **inspect_kwargs) > 0:
        sys.exit(1)


if __name__ == "__main__":
    better_exchook.install()
    main()



"""
Internal frontend for PyTorch frontend.
"""

from __future__ import annotations
from typing import Optional, Tuple

import torch

from returnn._internal_frontend_api import InternalFrontend
from returnn.tensor import Tensor

_TT = Tensor[torch.Tensor]


# Ignore this warning until we implemented everything.
# noinspection PyAbstractClass
class TorchInternalFrontend(InternalFrontend[torch.Tensor]):
    """
    Internal frontend for PyTorch frontend.
    """

    @staticmethod
    def get_dtype_name_raw(raw_tensor: torch.Tensor) -> str:
        """
        :return: dtype of raw tensor, as string
        """
        return str(raw_tensor.dtype).replace("torch.", "")

    @staticmethod
    def get_ndim_raw(raw_tensor: torch.Tensor) -> int:
        """
        :return: ndim of raw tensor
        """
        return raw_tensor.dim()

    @staticmethod
    def get_known_shape_raw(raw_tensor: torch.Tensor) -> Tuple[Optional[int]]:
        """
        :return: shape of raw tensor; here for PyTorch the full shape is always known
        """
        return tuple(raw_tensor.size())



"""
Here are all subprocess, threading etc related utilities,
most of them quite low level.
"""

from __future__ import annotations
from threading import Lock
import sys

import os
import io
import pickle
import types
import struct
import marshal
from importlib import import_module
import numpy

try:
    from _multiprocessing import Connection
except ImportError:
    from multiprocessing.connection import Connection

PY3 = sys.version_info[0] >= 3

if PY3:
    from io import BytesIO
else:
    # noinspection PyUnresolvedReferences,PyCompatibility
    from StringIO import StringIO as BytesIO


_abs_mod_file = os.path.abspath(__file__)


class SharedMem:
    class ShmException(Exception):
        pass

    class CCallException(ShmException):
        pass

    if sys.platform != "win32":
        import ctypes
        import ctypes.util

        libc_so = ctypes.util.find_library("c")
        libc = ctypes.CDLL(libc_so, use_errno=True)
        shm_key_t = ctypes.c_int
        IPC_PRIVATE = 0
        IPC_RMID = 0

        # int shmget(key_t key, size_t size, int shmflg);
        shmget = libc.shmget
        shmget.restype = ctypes.c_int
        shmget.argtypes = (shm_key_t, ctypes.c_size_t, ctypes.c_int)
        # void* shmat(int shmid, const void *shmaddr, int shmflg);
        shmat = libc.shmat
        shmat.restype = ctypes.c_void_p
        shmat.argtypes = (ctypes.c_int, ctypes.c_void_p, ctypes.c_int)
        # int shmdt(const void *shmaddr);
        shmdt = libc.shmdt
        shmdt.restype = ctypes.c_int
        shmdt.argtypes = (ctypes.c_void_p,)
        # int shmctl(int shmid, int cmd, struct shmid_ds *buf);
        shmctl = libc.shmctl
        shmctl.restype = ctypes.c_int
        shmctl.argtypes = (ctypes.c_int, ctypes.c_int, ctypes.c_void_p)
        # void* memcpy( void *dest, const void *src, size_t count );
        memcpy = libc.memcpy
        memcpy.restype = ctypes.c_void_p
        memcpy.argtypes = (ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t)

        @classmethod
        def check_ccall_error(cls, check, f):
            import ctypes

            if not check:
                errno = ctypes.get_errno()
                errstr = os.strerror(errno)
                raise cls.CCallException("SharedMem: %s failed with error %i (%s)" % (f, errno, errstr))

        @classmethod
        def is_shmget_functioning(cls):
            shmid = cls.shmget(cls.IPC_PRIVATE, 4 * 1024 * 1024, 0o600)
            if shmid <= 0:
                return False
            cls.shmctl(shmid, cls.IPC_RMID, 0)
            return True

        def __init__(self, size, shmid=None):
            self.size = size
            self.shmid = None
            self.ptr = None
            if shmid is None:
                self.is_creator = True
                self.shmid = self.shmget(self.IPC_PRIVATE, self.size, 0o600)
                self.check_ccall_error(self.shmid > 0, "shmget")
                print("SharedMem[pid %i]: New shmid: %i (size %i)" % (os.getpid(), self.shmid, self.size))
                import atexit

                atexit.register(self.remove)
            else:
                self.is_creator = False
                self.shmid = shmid
                assert self.shmid > 0
            self.ptr = self.shmat(self.shmid, 0, 0)
            self.check_ccall_error(self.ptr != self.ctypes.c_void_p(-1).value, "shmat")
            self.check_ccall_error(self.ptr > 0, "shmat")

        def remove(self):
            if self.ptr:
                self.shmdt(self.ptr)
                self.ptr = None
            if self.shmid and self.shmid > 0:
                if self.is_creator:
                    print("SharedMem[pid %i]: Removing shmid %i (size %i)" % (os.getpid(), self.shmid, self.size))
                    self.shmctl(self.shmid, self.IPC_RMID, 0)
                self.shmid = None

        def __del__(self):
            self.remove()

        def __getstate__(self):
            return {"size": self.size, "shmid": self.shmid}

        def __setstate__(self, state):
            self.__init__(**state)

        def __repr__(self):
            return "<SharedMem shmid=%r size=%r is_creator=%r>" % (self.shmid, self.size, self.is_creator)


def next_power_of_two(n):
    return 2 ** (int(n - 1).bit_length())


class SharedNumpyArray:
    """
    This class provides a way to create Numpy arrays in shared memory.
    It adds some logic to mark whether some shared memory segment can be reused
    - that is when the client marks it as unused.

    Note that there are a few similar Python modules:
      https://pypi.python.org/pypi/SharedArray
      https://parad0x.org/git/python/shared-array/about
      https://bitbucket.org/cleemesser/numpy-sharedmem/src
      https://stackoverflow.com/questions/5033799/how-do-i-pass-large-numpy-arrays
      https://stackoverflow.com/questions/7894791/use-numpy-array-in-shared-memory
    """

    # cls members
    ServerLock = Lock()
    ServerInstances = set()
    ServerArrayId = 0

    class TooMuchInstances(SharedMem.ShmException):
        pass

    ExtraSpaceBytes = 4096
    # local members
    is_server = False
    mem = None
    shape, strides, typestr = None, None, None

    @staticmethod
    def numpy_strides_for_fortran(shape, typestr):
        itemsize = numpy.dtype(typestr).itemsize
        strides = [itemsize]
        for s in shape:
            strides += [strides[-1] * s]
        strides = strides[:-1]
        return tuple(strides)

    @staticmethod
    def numpy_strides_for_c_contiguous(shape, typestr):
        itemsize = numpy.dtype(typestr).itemsize
        strides = [numpy.prod(shape[i + 1 :], dtype="int") * itemsize for i in range(len(shape))]
        return tuple(strides)

    @classmethod
    def needed_mem_size(cls, shape, typestr):
        itemsize = numpy.dtype(typestr).itemsize
        mem_size = cls.ExtraSpaceBytes + itemsize * numpy.prod(shape)
        return mem_size

    @classmethod
    def as_shared(cls, array):
        assert isinstance(array, numpy.ndarray)
        if isinstance(array.base, SharedNumpyArray):
            assert array.base.is_in_use()
            return array.base
        return cls.create_copy(array)

    @classmethod
    def create_copy(cls, array):
        assert isinstance(array, numpy.ndarray)
        array_intf = array.__array_interface__
        shape = array_intf["shape"]
        strides = array_intf["strides"]
        typestr = array_intf["typestr"]
        if array.flags.c_contiguous or array.flags.f_contiguous:
            pass  # ok, we can reuse it like that
        else:
            assert strides
            # Use some similar strides so that the copying might be faster.
            if strides[0] == array.itemsize:
                strides = cls.numpy_strides_for_fortran(shape=shape, typestr=typestr)
            else:
                strides = None  # C-contiguous
        inst = cls.create_new(shape=shape, strides=strides, typestr=typestr)
        inst.create_numpy_array()[...] = array
        assert inst._get_sanity_check_flag_ref().value == 42
        assert inst.is_in_use()
        return inst

    @classmethod
    def create_new(cls, shape, strides, typestr):
        needed_mem_size = cls.needed_mem_size(shape=shape, typestr=typestr)
        with cls.ServerLock:
            for inst in cls.ServerInstances:
                assert isinstance(inst, SharedNumpyArray)
                assert inst._get_sanity_check_flag_ref().value == 42
                if inst.is_in_use():
                    continue
                if inst.mem.size < needed_mem_size:
                    inst._init_mem(shape=shape, typestr=typestr)
                # We can reuse it.
                inst._set_new_array_id()
                inst._set_is_used(1)
                inst._set_numpy_format(shape=shape, strides=strides, typestr=typestr)
                return inst
        return cls(shape=shape, strides=strides, typestr=typestr)

    @classmethod
    def _get_new_array_id(cls):
        array_id = cls.ServerArrayId
        cls.ServerArrayId += 1
        return array_id

    def _set_new_array_id(self):
        assert self.is_server
        self.array_id = self._get_new_array_id()
        self._get_array_id_ref().value = self.array_id

    def __init__(self, shape, strides, typestr, mem=None, array_id=None):
        if not mem:
            assert array_id is None
            if len(self.ServerInstances) >= SharedMemNumpyConfig["max_server_instances"]:
                raise self.TooMuchInstances("too much instances (%i)" % len(self.ServerInstances))
            self.is_server = True
            self._init_mem(shape=shape, typestr=typestr)
            self._set_new_array_id()
            self._set_is_used(1)
        else:
            assert array_id is not None
            self.is_server = False
            self.array_id = array_id
            mem_size = self.needed_mem_size(shape=shape, typestr=typestr)
            assert isinstance(mem, SharedMem)
            assert mem.size >= mem_size
            assert mem.shmid > 0
            assert mem.ptr > 0
            self.mem = mem
            assert self._get_sanity_check_flag_ref().value == 42
            assert self._get_array_id_ref().value == self.array_id
            assert self.is_in_use()
        self._set_numpy_format(shape=shape, strides=strides, typestr=typestr)
        if self.is_server:
            with self.ServerLock:
                self.ServerInstances.add(self)

    def _set_numpy_format(self, shape, strides, typestr):
        itemsize = numpy.dtype(typestr).itemsize
        if strides:
            assert all([st > 0 for st in strides])
            assert sum([st * (sh - 1) for (st, sh) in zip(strides, shape)]) + itemsize == numpy.prod(shape) * itemsize
        self.shape = shape
        self.strides = strides
        self.typestr = typestr

    def _init_mem(self, shape, typestr):
        assert self.is_server
        if self.mem:
            self.mem.remove()
            self.mem = None
        assert numpy.prod(shape) > 0
        mem_size = next_power_of_two(self.needed_mem_size(shape=shape, typestr=typestr))
        mem_size = max(SharedMemNumpyConfig["min_shared_mem_size"], mem_size)
        self.mem = SharedMem(size=mem_size)
        self._get_sanity_check_flag_ref().value = 42

    def get_numpy_array_data_ptr(self):
        assert self.mem.ptr > 0
        return self.mem.ptr + self.ExtraSpaceBytes

    @property
    def __array_interface__(self):
        assert self.shape
        # https://docs.scipy.org/doc/numpy/reference/arrays.interface.html
        return {
            "data": (self.get_numpy_array_data_ptr(), False),
            "shape": self.shape,
            "strides": self.strides,
            "typestr": self.typestr,
            "version": 3,
        }

    def create_numpy_array(self):
        assert self._get_sanity_check_flag_ref().value == 42
        assert self._get_array_id_ref().value == self.array_id
        assert self.is_in_use()
        a = numpy.array(self, copy=False)
        assert a.__array_interface__["data"][0] == self.get_numpy_array_data_ptr()
        assert not a.flags.owndata, "a.__array_interface__ = %r" % a.__array_interface__
        assert a.base is self
        assert a.nbytes + self.ExtraSpaceBytes <= self.mem.size
        assert (
            sum([st * (sh - 1) for (st, sh) in zip(a.strides, a.shape)]) + a.itemsize
            == numpy.prod(a.shape) * a.itemsize
            == a.nbytes
        )
        return a

    def _get_sanity_check_flag_ref(self):
        assert self.mem.ptr > 0
        import ctypes

        return ctypes.cast(ctypes.c_void_p(self.mem.ptr), ctypes.POINTER(ctypes.c_uint64)).contents

    def _get_array_id_ref(self):
        assert self.mem.ptr > 0
        import ctypes

        return ctypes.cast(ctypes.c_void_p(self.mem.ptr + 8), ctypes.POINTER(ctypes.c_uint64)).contents

    def _get_in_use_flag_ref(self):
        assert self.mem.ptr > 0
        import ctypes

        return ctypes.cast(ctypes.c_void_p(self.mem.ptr + 16), ctypes.POINTER(ctypes.c_uint64)).contents

    def _set_is_used(self, n):
        self._get_in_use_flag_ref().value = n

    def is_in_use(self):
        return self._get_in_use_flag_ref().value > 0

    def set_unused(self):
        if self.is_server:
            return
        if self.mem:
            self._set_is_used(0)
            self.mem.remove()
            self.mem = None

    def __getstate__(self):
        return {
            "shape": self.shape,
            "strides": self.strides,
            "typestr": self.typestr,
            "mem": self.mem,
            "array_id": self.array_id,
        }

    def __setstate__(self, state):
        self.__init__(**state)

    def __del__(self):
        # On the server side, we will get deleted at program end
        # because we are referenced in the global SharedNumpyArray.ServerInstances.
        # On the client side, we will get deleted once we are not used anymore.
        # Note that self.array holds a reference to self.
        self.set_unused()

    def __repr__(self):
        return "<%s is_server=%r state=%r>" % (self.__class__.__name__, self.is_server, self.__getstate__())


def attrChain(base, *attribs, **kwargs):
    default = kwargs.get("default", None)
    obj = base
    for attr in attribs:
        if obj is None:
            return default
        obj = getattr(obj, attr, None)
    if obj is None:
        return default
    return obj


# This is needed in some cases to avoid pickling problems with bounded funcs.
def funcCall(attrChainArgs, args=()):
    f = attrChain(*attrChainArgs)
    return f(*args)


Unpickler = pickle.Unpickler
if PY3:

    def get_func_closure(f):
        return f.__closure__

    # (code, globals[, name[, argdefs[, closure]]])
    def get_func_tuple(f):
        return (
            f.__code__,
            f.__globals__,
            f.__name__,
            f.__defaults__,
            f.__closure__,
        )

else:

    def get_func_closure(f):
        return f.func_closure

    def get_func_tuple(f):
        return (
            f.func_code,
            f.func_globals,
            f.func_name,
            f.func_defaults,
            f.func_closure,
        )


_closure = (lambda x: lambda: x)(0)
# noinspection PyUnresolvedReferences
_cell = get_func_closure(_closure)[0]
CellType = type(_cell)
ModuleType = type(sys)
# noinspection PyUnresolvedReferences
DictType = dict if PY3 else types.DictionaryType

if PY3:

    class BufferType:
        "Dummy"

    def make_buffer(*args):
        assert False

else:
    # noinspection PyUnresolvedReferences
    make_buffer = buffer
    # noinspection PyUnresolvedReferences
    BufferType = types.BufferType

    def bytes(x, *args):
        return str(x)


if PY3:
    _old_style_class = None

    class OldStyleClass:
        "Dummy"

    class _new_style_class:
        pass

    NewStyleClass = type
else:

    class _old_style_class:
        pass

    class _new_style_class(object):
        pass

    OldStyleClass = type(_old_style_class)  # == types.ClassType (classobj)
    NewStyleClass = type(_new_style_class)  # (type)


def makeFuncCell(value):
    return get_func_closure((lambda: value))[0]


def getModuleDict(modname, path=None):
    """
    :param str modname: such that "import <modname>" would work
    :param list[str] path: sys.path
    :return: the dict of the mod
    :rtype: dict[str]
    """
    try:
        mod = import_module(modname)
    except ImportError:
        # Try again with extended sys.path.
        assert path
        for p in path:
            if p not in sys.path:
                sys.path.append(p)
        mod = import_module(modname)
    return mod.__dict__


def getModNameForModDict(obj):
    """
    :type obj: dict
    :rtype: str | None
    :returns The module name or None. It will not return '__main__' in any case
    because that likely will not be the same in the unpickling environment.
    Also see: https://stackoverflow.com/questions/56171796/
    """
    if "__name__" not in obj:
        return None  # this is not a module
    mod_name = obj["__name__"]
    if mod_name == "__main__":
        return None
    if mod_name not in sys.modules:
        return None  # does not look like we have it loaded
    mod = sys.modules[mod_name]
    if mod.__dict__ is obj:
        return mod_name
    return None


def getNormalDict(d):
    """
    :type d: dict[str] | dictproxy
    :rtype: dict[str]
    It also removes getset_descriptor. New-style classes have those.
    """
    r = {}
    for k, v in d.items():
        if isinstance(v, types.GetSetDescriptorType):
            continue
        r[k] = v
    return r


def make_numpy_ndarray_fromstring(s, dtype, shape):
    return numpy.fromstring(s, dtype=dtype).reshape(shape)


SharedMemNumpyConfig = {
    "enabled": False,
    "auto_pickling_min_size": 8 * 1024 * 1024,  # 8MB
    "min_shared_mem_size": 32 * 1024 * 1024,  # 32MB
    "max_server_instances": 10,
}


def use_shared_mem_for_numpy_array(obj):
    assert isinstance(obj, numpy.ndarray)
    if obj.shape == ():  # scalar
        return False  # cannot use shared memory because it will always use its own memory
    if isinstance(obj.base, SharedNumpyArray):
        assert obj.base.is_in_use()
        return True
    if not SharedMemNumpyConfig["enabled"]:
        return False
    return obj.nbytes >= SharedMemNumpyConfig["auto_pickling_min_size"]


def numpy_set_unused(v):
    """
    :param numpy.ndarray v: array which will be marked as not-used-anymore
    This will tell mechanisms like SharedNumpyArray that it can reuse the memory.
    On the client side, this will even unmap the memory, so any further access
    to it will cause a SEGFAULT.
    """
    if v is None:
        return
    assert isinstance(v, numpy.ndarray)
    if isinstance(v.base, SharedNumpyArray):
        assert v.base.is_in_use()  # must not be called multiple times
        v.base.set_unused()


def numpy_copy_and_set_unused(v):
    """
    :param dict[str,numpy.ndarray|object] | numpy.ndarray | object v: object to be handled
    If v is a dict, we will return a new copied dict where every value is mapped through numpy_copy_and_set_unused.
    If v is a numpy.ndarray and its base is a SharedNumpyArray, we will copy it and
      call numpy_set_unused on the old value.
    If v is a numpy.ndarray and its base is not a SharedNumpyArray, we will just return it as it is and do nothing.
    In all other cases, we will also just return the object as it is and do nothing.
    """
    if isinstance(v, numpy.ndarray):
        if isinstance(v.base, SharedNumpyArray):
            newv = v.copy(order="A")
            numpy_set_unused(v)
            return newv
        return v
    if isinstance(v, dict):
        return {k: numpy_copy_and_set_unused(vv) for (k, vv) in v.items()}
    return v


def numpy_alloc(shape, dtype, fortran_for_shared=False):
    """
    If EnableAutoNumpySharedMemPickling is True, this will allocate a Numpy array
    in shared memory so we avoid a copy later on when this Numpy array would
    be transferred to another process via pickling.
    """
    if SharedMemNumpyConfig["enabled"]:
        dtype = numpy.dtype(dtype)
        typestr = dtype.str
        strides = None
        if fortran_for_shared:
            strides = SharedNumpyArray.numpy_strides_for_fortran(shape=shape, typestr=typestr)
        try:
            return SharedNumpyArray.create_new(shape=shape, strides=strides, typestr=typestr)
        except SharedMem.ShmException as e:
            print("numpy_alloc: SharedMem exception: %s" % e)
    # Fallback.
    return numpy.ndarray(shape, dtype=dtype)


try:
    _BasePickler = pickle._Pickler  # use the pure Python implementation
except AttributeError:
    _BasePickler = pickle.Pickler


class Pickler(_BasePickler):
    """
    We extend the standard Pickler to be able to pickle some more types,
    such as lambdas and functions, code, func cells, buffer and more.
    """

    def __init__(self, *args, **kwargs):
        if not "protocol" in kwargs:
            kwargs["protocol"] = pickle.HIGHEST_PROTOCOL
        _BasePickler.__init__(self, *args, **kwargs)

    dispatch = _BasePickler.dispatch.copy()

    def save_func(self, obj):
        try:
            self.save_global(obj)
            return
        except pickle.PicklingError:
            pass
        assert type(obj) is types.FunctionType
        self.save(types.FunctionType)
        self.save(get_func_tuple(obj))
        self.write(pickle.REDUCE)
        if id(obj) not in self.memo:  # Could be if we recursively landed here. See also pickle.save_tuple().
            self.memoize(obj)

    dispatch[types.FunctionType] = save_func

    def save_method(self, obj):
        try:
            self.save_global(obj)
            return
        except pickle.PicklingError:
            pass
        assert type(obj) is types.MethodType
        self.save(types.MethodType)
        if PY3:
            self.save((obj.__func__, obj.__self__))
        else:
            self.save((obj.im_func, obj.im_self, obj.im_class))
        self.write(pickle.REDUCE)
        self.memoize(obj)

    dispatch[types.MethodType] = save_method

    def save_code(self, obj):
        assert type(obj) is types.CodeType
        self.save(marshal.loads)
        self.save((marshal.dumps(obj),))
        self.write(pickle.REDUCE)
        self.memoize(obj)

    dispatch[types.CodeType] = save_code

    def save_cell(self, obj):
        assert type(obj) is CellType
        self.save(makeFuncCell)
        self.save((obj.cell_contents,))
        self.write(pickle.REDUCE)
        self.memoize(obj)

    dispatch[CellType] = save_cell

    # We also search for module dicts and reference them.
    # This is for FunctionType.func_globals.
    def intellisave_dict(self, obj):
        modname = getModNameForModDict(obj)
        if modname:
            self.save(getModuleDict)
            self.save((modname, sys.path))
            self.write(pickle.REDUCE)
            self.memoize(obj)
            return
        self.save_dict(obj)

    dispatch[DictType] = intellisave_dict

    def save_module(self, obj):
        modname = getModNameForModDict(obj.__dict__)
        if modname:
            self.save(import_module)
            self.save((modname,))
            self.write(pickle.REDUCE)
            self.memoize(obj)
            return
        # We could maybe construct it manually. For now, just fail.
        raise pickle.PicklingError("cannot pickle module %r" % obj)

    dispatch[ModuleType] = save_module

    def save_buffer(self, obj):
        self.save(buffer)
        self.save((str(obj),))
        self.write(pickle.REDUCE)

    dispatch[BufferType] = save_buffer

    def save_string(self, obj, pack=struct.pack):
        # Difference to base: We just always use BINSTRING (simpler)
        # and use a separate write for the obj itself.
        # For a huge obj, this avoids one unnecessary copy of the data.
        self.write(pickle.BINSTRING + pack("<i", len(obj)))
        self.write(bytes(obj, "utf8"))

    dispatch[str] = save_string

    def save_ndarray(self, obj):
        if use_shared_mem_for_numpy_array(obj):
            try:
                shared = SharedNumpyArray.as_shared(obj)
            except SharedMem.ShmException as e:
                print("SharedNumpyArray exception: %s" % e)
                # fallback to default
            else:
                self.save(shared.create_numpy_array)
                self.save(())
                self.write(pickle.REDUCE)
                return
        # For some reason, Numpy fromstring/tostring is faster than Numpy loads/dumps.
        self.save(make_numpy_ndarray_fromstring)
        self.save((obj.tostring(), str(obj.dtype), obj.shape))
        self.write(pickle.REDUCE)

    dispatch[numpy.ndarray] = save_ndarray

    def save_iobuffer_dummy(self, obj):
        # Not supported but we want to not fail and just store None.
        self.save_none(None)

    dispatch[io.BufferedReader] = save_iobuffer_dummy
    dispatch[io.BufferedWriter] = save_iobuffer_dummy

    # Overwrite to avoid the broken pickle.whichmodule() which might return "__main__".
    def save_global(self, obj, name=None):
        assert obj
        assert id(obj) not in self.memo
        if name is None:
            name = obj.__name__

        module = getattr(obj, "__module__", None)
        if module == "__main__" and globals().get(name, None) is obj:
            # Can happen if this is directly executed.
            module = __name__  # should be correct now
        if module is None or module == "__main__":
            module = pickle.whichmodule(obj, name)
        if module is None or module == "__main__":
            raise pickle.PicklingError("Can't pickle %r: module not found: %s" % (obj, module))

        try:
            __import__(module)
            mod = sys.modules[module]
            klass = getattr(mod, name)
        except (ImportError, KeyError, AttributeError):
            raise pickle.PicklingError("Can't pickle %r: it's not found as %s.%s" % (obj, module, name))
        else:
            if klass is not obj:
                raise pickle.PicklingError("Can't pickle %r: it's not the same object as %s.%s" % (obj, module, name))

        assert "\n" not in module
        assert "\n" not in name
        self.write(pickle.GLOBAL + bytes(module + "\n" + name + "\n", "utf8"))
        self.memoize(obj)

    def save_type(self, obj):
        try:
            self.save_global(obj)
            return
        except pickle.PicklingError:
            pass
        # Some types in the types modules are not correctly referenced,
        # such as types.FunctionType. This is fixed here.
        for modname in ["types"]:
            moddict = sys.modules[modname].__dict__
            for modobjname, modobj in moddict.items():
                if modobj is obj:
                    self.write(pickle.GLOBAL + bytes(modname + "\n" + modobjname + "\n", "utf8"))
                    self.memoize(obj)
                    return
        # Generic serialization of new-style classes.
        self.save(type)
        self.save((obj.__name__, obj.__bases__, getNormalDict(obj.__dict__)))
        self.write(pickle.REDUCE)
        self.memoize(obj)

    dispatch[NewStyleClass] = save_type

    # This is about old-style classes.
    def save_class(self, cls):
        try:
            # First try with a global reference. This works normally. This is the default original pickle behavior.
            self.save_global(cls)
            return
        except pickle.PicklingError:
            pass
        # It didn't worked. But we can still serialize it.
        # Note that this could potentially confuse the code if the class is reference-able in some other way
        # - then we will end up with two versions of the same class.
        self.save(types.ClassType)
        self.save((cls.__name__, cls.__bases__, cls.__dict__))
        self.write(pickle.REDUCE)
        self.memoize(cls)
        return

    dispatch[OldStyleClass] = save_class

    # avoid pickling instances of ourself. this mostly doesn't make sense and leads to trouble.
    # however, also doesn't break. it mostly makes sense to just ignore.
    def __getstate__(self):
        return None

    def __setstate__(self, state):
        pass



"""
The :class:`Tensor` class represents a tensor.
This is framework-agnostic, i.e. is fine for TensorFlow, PyTorch, or any other framework.
(Earlier, this was the class ``Data`` in RETURNN, and ``Data`` is now an alias for this class.)

This class is to describe a tensor,
i.e. its shape and properties like
whether we should consider it sparse data (i.e. it represents indices).
Each dimension is described by :class:`Dim`.

This is used in :class:`TFNetwork` to describe the dataset external data (:class:`ExternData`)
as well as in every layer's output and in many other parts of the code.

See :ref:`data`.

This is designed in a way that is efficient for eager-mode execution such as PyTorch,
but at the same time compatible with older RETURNN code.

Discussion on the move from loosely TF-specific ``Data`` to framework-agnostic ``Tensor``:
https://github.com/rwth-i6/returnn/issues/1165
"""

from __future__ import annotations
from typing import Optional, Sequence, Tuple, Generic, TypeVar

from .dim import Dim
from ._tensor_extra import _TensorExtra, _TensorMixin
from . import _tensor_extra

RawTensorType = TypeVar("RawTensorType")  # e.g. torch.Tensor, tf.Tensor, numpy.ndarray, ...


class Tensor(_TensorMixin, Generic[RawTensorType]):
    """
    Represents a tensor, in a frame-agnostic way. See the module docstring.
    """

    size_dtype = "int32"

    __slots__ = ("name", "_dims", "dtype", "sparse_dim", "_raw_tensor", "version", "_extra")

    name: str
    _dims: Tuple[Dim, ...]
    dtype: str
    sparse_dim: Optional[Dim]
    _raw_tensor: Optional[RawTensorType]
    version: int
    _extra: Optional[_TensorExtra]

    def __init__(
        self,
        name: str,
        dims: Optional[Sequence[Dim]] = None,
        dtype: Optional[str] = None,
        *,
        sparse_dim: Optional[Dim] = None,
        raw_tensor: Optional[RawTensorType] = None,
        version: Optional[int] = None,
        **kwargs,
    ):
        """
        :param name:
        :param dims: the shape, where each dimension is described by a :class:`Dim`.
        :param dtype: e.g. "float32" or "int64"
        :param sparse_dim: when the values are indices into some dimension, this is the dimension.
            You can also interpret the whole tensor as a sparse representation of a dense one-hot tensor,
            where this sparse_dim becomes the additional dense dimension.
        :param raw_tensor: the raw tensor, e.g. numpy array, TF tensor, or PyTorch tensor
        :param version: behavior version just for Tensor. If not specified, and `dims` is None (old code),
            it uses version 1.
            - v1: the old behavior of Data. Specifically, time_dim_axis and feature_dim_axis are used
                and automatically inferred when not specified.
            - v2: time_dim_axis, feature_dim_axis are None by default.
        :param kwargs: see :func:`_handle_extra_kwargs`, :func:`infer_dim_tags`
        """
        if "sparse" in kwargs and sparse_dim is None:
            sparse_dim = _tensor_extra.infer_sparse_dim(name=name, sparse_dim=sparse_dim, **kwargs)
        if dims is not None:
            assert "shape" not in kwargs and "dim_tags" not in kwargs  # probably old code got this wrong
            if version is None:
                version = 2
        else:
            # old code
            dims = _tensor_extra.infer_dim_tags(name=name, sparse_dim=sparse_dim, **kwargs)
            if version is None:
                version = 1
        if dtype is None:
            # old defaults
            dtype = "int32" if sparse_dim else "float32"

        self.name = name
        self._dims = tuple(dims)
        self.dtype = dtype
        self.sparse_dim = sparse_dim
        self._raw_tensor = None  # assignment below
        self.version = version
        self._extra = None  # type: Optional[_TensorExtra]

        if kwargs:
            self._handle_extra_kwargs(**kwargs)
        if raw_tensor is not None:
            self.raw_tensor = raw_tensor  # assignment via property, to have extra checks

    @property
    def dims(self) -> Tuple[Dim, ...]:
        """
        :return: dim tags
        """
        return self._dims

    @property
    def raw_tensor(self) -> Optional[RawTensorType]:
        """
        :return: raw tensor
        """
        return self._raw_tensor

    @raw_tensor.setter
    def raw_tensor(self, value: Optional[RawTensorType]):
        """
        :param value:
        """
        self._raw_tensor = value
        self.sanity_check(assume_complete=False)


# dispatch table for frameworks for sanity_check



"""
Internal frontend for TF low-level frontend.
"""

from __future__ import annotations
from typing import Optional, Union, Any, Sequence, Tuple
import tensorflow as tf
import contextlib

from returnn._internal_frontend_api import InternalFrontend
from returnn.tensor import Tensor
from returnn.tf.util import basic as tf_util
from returnn.tf import compat as tf_compat

_TT = Tensor[tf.Tensor]


class TFInternalFrontend(InternalFrontend[tf.Tensor]):
    """
    Internal frontend for TF low-level frontend.
    """

    @staticmethod
    def get_dtype_name_raw(raw_tensor: tf.Tensor) -> str:
        """
        :return: dtype of raw tensor, as string
        """
        return raw_tensor.dtype.base_dtype.name

    @staticmethod
    def get_ndim_raw(raw_tensor: tf.Tensor) -> int:
        """
        :return: ndim of raw tensor. assumes it is known
        """
        assert raw_tensor.shape.ndims is not None
        return raw_tensor.shape.ndims

    @staticmethod
    def get_shape_raw(raw_tensor: tf.Tensor) -> tf.Tensor:
        """
        :return: shape of raw tensor
        """
        return tf.shape(raw_tensor)

    @staticmethod
    def get_shape_tuple_raw(raw_tensor: tf.Tensor) -> Tuple[Union[int, tf.Tensor]]:
        """
        :return: shape of raw tensor. assumes that ndim is known
        """
        shape = raw_tensor.shape.as_list()
        if all([dim is not None for dim in shape]):
            return tuple(shape)
        with tf_util.same_control_flow_ctx(raw_tensor):
            shape_dynamic = tf.shape(raw_tensor)
            for axis, dim in enumerate(shape):
                if dim is None:
                    shape[axis] = shape_dynamic[axis]
            return tuple(shape)

    @staticmethod
    def get_known_shape_raw(raw_tensor: tf.Tensor) -> Tuple[Optional[int]]:
        """
        :return: shape of raw tensor, int for static known, None otherwise. assumes that ndim is known.
        """
        return tuple(raw_tensor.shape.as_list())

    @staticmethod
    def set_known_shape_raw(raw_tensor: tf.Tensor, shape: Tuple[Optional[int]]) -> None:
        """
        wrap tf.Tensor.set_shape
        """
        raw_tensor.set_shape(shape)

    @staticmethod
    def fill_raw(shape: Union[Sequence[Union[int, tf.Tensor]], tf.Tensor], value: Union[Any, tf.Tensor]) -> tf.Tensor:
        """
        :param shape: shape
        :param value: value to fill
        :return: raw tensor filled with value everywhere
        """
        with tf_util.same_control_flow_ctx([shape, value]):
            return tf.fill(shape, value)

    @staticmethod
    def compare_raw(a: tf.Tensor, kind: str, b: tf.Tensor) -> tf.Tensor:
        """
        :param a:
        :param kind: "equal"|"==", "less"|"<", "less_equal"|"<=", "greater"|">", "greater_equal"|">=", "not_equal"|"!="
        :param b:
        :return: a `kind` b
        """
        assert a.shape.ndims == b.shape.ndims
        kind = {
            "==": "equal",
            "<=": "less_equal",
            "<": "less",
            ">=": "greater_equal",
            ">": "greater",
            "!=": "not_equal",
            "<>": "not_equal",
        }.get(kind, kind)
        op = getattr(tf, kind)  # e.g. tf.equal
        return op(a, b)

    @staticmethod
    def reshape_raw(raw_tensor: tf.Tensor, shape: Union[Sequence[Union[int, tf.Tensor]], tf.Tensor]) -> tf.Tensor:
        """
        :param raw_tensor: raw tensor
        :param shape: new shape
        :return: reshaped raw tensor
        """
        with tf_util.same_control_flow_ctx([raw_tensor, shape]):
            return tf.reshape(raw_tensor, shape)

    @classmethod
    def squeeze_raw(cls, raw_tensor: tf.Tensor, axes: Sequence[int]) -> tf.Tensor:
        """
        :param raw_tensor: raw tensor
        :param axes: axes to squeeze
        :return: squeezed raw tensor
        """
        known_shape = raw_tensor.shape.as_list()
        assert all([known_shape[axis] == 1 for axis in axes])
        with tf_util.same_control_flow_ctx(raw_tensor):
            return tf.squeeze(raw_tensor, axis=axes)

    @staticmethod
    def transpose_raw(raw_tensor: tf.Tensor, perm: Sequence[int]) -> tf.Tensor:
        """
        :param raw_tensor:
        :param perm: e.g. [0, 2, 1]
        :return: permuted (transposed) raw tensor; wraps tf.transpose
        """
        with tf_util.same_control_flow_ctx(raw_tensor):
            return tf.transpose(raw_tensor, perm)

    @staticmethod
    def expand_dims_raw(raw_tensor: tf.Tensor, axis: int) -> tf.Tensor:
        """
        :param raw_tensor:
        :param axis: e.g. 1
        :return: raw tensor with new axis; wraps tf.expand_dims
        """
        with tf_util.same_control_flow_ctx(raw_tensor):
            return tf.expand_dims(raw_tensor, axis=axis)

    @staticmethod
    def expand_raw(raw_tensor: tf.Tensor, axis: int, dim: Union[int, tf.Tensor]) -> tf.Tensor:
        """
        :param raw_tensor:
        :param axis: shape[axis] must be 1
        :param dim: the new dim for shape[axis]
        :return: shape[axis] expands to dim
        """
        assert raw_tensor.shape.as_list()[axis] == 1
        with tf_util.same_control_flow_ctx(raw_tensor):
            return tf.tile(raw_tensor, [1] * axis + [dim] + [1] * (raw_tensor.shape.ndims - axis - 1))

    @staticmethod
    def sequence_mask_raw(lengths: tf.Tensor, *, batch_major: bool = True) -> tf.Tensor:
        """
        Wraps around tf.sequence_mask().
        It will cache the value inside the passed object so that we don't recompute it multiple times.

        :param lengths: shape (batch,)
        :param batch_major:
        :return: tensor mask of shape (batch,maxlen) if batch_major else (maxlen,batch) of type bool
        """
        if batch_major:
            return tf_util.sequence_mask(lengths)
        else:
            return tf_util.sequence_mask_time_major(lengths)

    @staticmethod
    @contextlib.contextmanager
    def name_scope_raw(name: str) -> Any:
        """
        :param name: name of scope
        :return: context manager
        """
        with tf.name_scope(name):
            yield

    @staticmethod
    @contextlib.contextmanager
    def control_dependencies_raw(dependencies: Sequence[Union[tf.Tensor, tf.Operation]]) -> Any:
        """
        :param dependencies: list of tensors or operations
        :return: context manager
        """
        with tf.control_dependencies(dependencies):
            yield

    @staticmethod
    def identity_with_control_dependencies_raw(raw_tensor: tf.Tensor, dependencies: Sequence[Any]) -> tf.Tensor:
        """
        :param raw_tensor:
        :param dependencies: list of tensors or operations
        :return: identity of tensor with control dependencies
        """
        with tf.control_dependencies(dependencies), tf_util.same_control_flow_ctx(raw_tensor):
            return tf.identity(raw_tensor)

    @staticmethod
    def create_placeholder(tensor: _TT) -> tf.Tensor:
        """
        :return: tf.placeholder in TF
        """
        with tf.name_scope("extern_data/placeholders/%s/" % tensor.name):
            return tf_compat.v1.placeholder(**tensor.get_placeholder_kwargs(with_batch=True))

    @staticmethod
    def runtime_sanity_checks(tensor: _TT) -> tf.Operation:
        """
        Runtime checks
        """
        checks = []
        with tf.name_scope("runtime_sanity_check"):
            shape = tf.shape(tensor.placeholder)
            # noinspection PyShadowingNames
            batch_dim = shape[tensor.batch_dim_axis] if tensor.have_batch_axis() else 1
            rank = tf.rank(tensor.placeholder)
            data = ["Data.get_runtime_sanity_check_op:", str(tensor), "shape", shape]
            for i, tag in enumerate(tensor.dim_tags):
                if tag.dyn_size is not None:
                    data += ["dyn_size[%i] (%s)" % (i, tag), tag.dyn_size, ".shape", tf.shape(tag.dyn_size)]
            checks += [tf.Assert(tf.equal(rank, tensor.batch_ndim), data + ["-> invalid rank"])]
            if tensor.have_batch_axis():
                batch_dim_via_info = tensor.get_batch_dim()
                checks += [
                    tf.Assert(
                        tf.equal(batch_dim, batch_dim_via_info),
                        data + ["-> invalid batch dim info", batch_dim_via_info],
                    )
                ]
            for i in range(tensor.batch_ndim):
                if tensor.batch_shape[i] is not None:
                    checks += [
                        tf.Assert(tf.equal(shape[i], tensor.batch_shape[i]), data + ["-> invalid shape[%i]" % i])
                    ]
                dyn_size_ext = tensor.dim_tags[i].dyn_size_ext
                if dyn_size_ext and dyn_size_ext.placeholder is not None:
                    dyn_size = dyn_size_ext.placeholder
                    if dyn_size_ext.have_batch_axis() and tensor.have_batch_axis():
                        checks += [
                            tf.Assert(
                                tf.equal(tf.shape(dyn_size)[dyn_size_ext.batch_dim_axis], batch_dim),
                                data + ["-> invalid axis %i tag dyn size batch dim" % i],
                            )
                        ]
                    checks += [
                        tf.Assert(
                            # Note: in almost all cases, we have equality here.
                            # However, not strictly in all cases, e.g. DecideLayer, maybe some others...
                            # But that should not be more than 1 less.
                            tf.logical_or(
                                tf.logical_and(
                                    tf.less_equal(tf.reduce_max(dyn_size), shape[i]),
                                    tf.greater_equal(tf.reduce_max(dyn_size), shape[i] - 1),
                                ),
                                # In other rare cases, this might be a broadcast dim
                                # (e.g. as initial values of att weights for a rec loop).
                                tf.equal(1, shape[i]),
                            ),
                            data + ["-> invalid shape[%i] or max(dyn_size[%i])" % (i, i)],
                        )
                    ]
                    checks += [dyn_size_ext.get_runtime_sanity_check_op()]
        return tf.group(*checks)

    @staticmethod
    def is_valid_in_current_graph(tensor: _TT) -> bool:
        """
        :return: whether the tensor is valid in the current graph
        """
        if tensor.raw_tensor is None:
            return True
        if tf_compat.executing_eagerly():
            return True  # always true in eager mode
        g = tf_util.get_root_graph()
        return tf_util.get_root_graph(tensor.raw_tensor.graph) is g

    @staticmethod
    def format_graph_output(raw_tensor: tf.Tensor, *, max_depth: Optional[int] = None) -> str:
        """
        :param raw_tensor:
        :param max_depth:
        """
        return tf_util.format_graph_output(raw_tensor, max_depth=max_depth)



from __future__ import annotations

import _setup_test_env  # noqa
import sys
import os
import threading
from returnn.util.task_system import *
import unittest
import gc

SharedMemNumpyConfig["enabled"] = True
SharedMemNumpyConfig["auto_pickling_min_size"] = 1

try:
    # noinspection PyCompatibility
    from StringIO import StringIO
except ImportError:  # Python 3
    from io import BytesIO as StringIO


def pickle_dumps(obj):
    sio = StringIO()
    p = Pickler(sio)
    p.dump(obj)
    return sio.getvalue()


def pickle_loads(s):
    p = Unpickler(StringIO(s))
    return p.load()


def find_numpy_shared_by_shmid(shmid):
    for sh in SharedNumpyArray.ServerInstances:
        assert isinstance(sh, SharedNumpyArray)
        assert sh.mem is not None
        assert sh.mem.shmid > 0
        if sh.mem.shmid == shmid:
            return sh
    return None


_have_working_shmget = None


def have_working_shmget():
    global _have_working_shmget
    if _have_working_shmget is None:
        _have_working_shmget = SharedMem.is_shmget_functioning()
    print("shmget functioning:", _have_working_shmget)
    return _have_working_shmget


@unittest.skipIf(not have_working_shmget(), "shmget does not work")
def test_shmget_functioning():
    assert SharedMem.is_shmget_functioning()


@unittest.skipIf(not have_working_shmget(), "shmget does not work")
def test_pickle_numpy():
    m = numpy.random.randn(10, 10)
    p = pickle_dumps(m)
    m2 = pickle_loads(p)
    assert isinstance(m2, numpy.ndarray)
    assert numpy.allclose(m, m2)
    assert isinstance(m2.base, SharedNumpyArray)
    shared_client = m2.base
    assert not shared_client.is_server
    shared_server = find_numpy_shared_by_shmid(m2.base.mem.shmid)
    assert shared_server.is_server
    assert numpy.allclose(m, shared_server.create_numpy_array())
    assert numpy.allclose(m, shared_client.create_numpy_array())
    assert shared_server.is_in_use()
    assert shared_client.is_in_use()
    numpy_set_unused(m2)
    assert not shared_server.is_in_use()
    assert shared_client.mem is None


@unittest.skipIf(not have_working_shmget(), "shmget does not work")
def test_pickle_numpy_scalar():
    # Note that a scalar really does not work because numpy.array(float) will always own its data.
    m = numpy.array([numpy.random.randn()])
    assert isinstance(m, numpy.ndarray)
    assert m.shape == (1,)
    assert m.nbytes >= 1
    p = pickle_dumps(m)
    m2 = pickle_loads(p)
    assert isinstance(m2, numpy.ndarray)
    assert numpy.allclose(m, m2)
    assert isinstance(m2.base, SharedNumpyArray)
    shared_client = m2.base
    assert not shared_client.is_server
    shared_server = find_numpy_shared_by_shmid(m2.base.mem.shmid)
    assert shared_server.is_server
    assert numpy.allclose(m, shared_server.create_numpy_array())
    assert numpy.allclose(m, shared_client.create_numpy_array())
    assert shared_server.is_in_use()
    assert shared_client.is_in_use()
    numpy_set_unused(m2)
    assert not shared_server.is_in_use()
    assert shared_client.mem is None


@unittest.skipIf(not have_working_shmget(), "shmget does not work")
def test_pickle_gc_aggressive():
    m = numpy.random.randn(10, 10)
    p = pickle_dumps(m)
    m2 = pickle_loads(p)
    assert isinstance(m2, numpy.ndarray)
    assert numpy.allclose(m, m2)
    assert isinstance(m2.base, SharedNumpyArray)
    print("refcount: %i" % sys.getrefcount(m2.base))
    gc.collect()
    gc.collect()
    print("refcount: %i" % sys.getrefcount(m2.base))
    assert m2.base.is_in_use()
    server = find_numpy_shared_by_shmid(m2.base.mem.shmid)
    m2 = None
    gc.collect()


@unittest.skipIf(not have_working_shmget(), "shmget does not work")
def test_pickle_multiple():
    for i in range(20):
        ms = [numpy.random.randn(10, 10) for i in range(i % 3 + 1)]
        p = pickle_dumps(ms)
        ms2 = pickle_loads(p)
        assert len(ms) == len(ms2)
        for m, m2 in zip(ms, ms2):
            assert numpy.allclose(m, m2)
            assert isinstance(m2.base, SharedNumpyArray)


@unittest.skipIf(not have_working_shmget(), "shmget does not work")
def test_pickle_unpickle_auto_unused():
    old_num_servers = None
    for i in range(10):
        m = numpy.random.randn(i * 2 + 1, i * 3 + 2)
        p = pickle_dumps((m, m, m))
        new_num_servers = len(SharedNumpyArray.ServerInstances)
        if old_num_servers is not None:
            assert old_num_servers == new_num_servers
        old_num_servers = new_num_servers
        m2, m3, m4 = pickle_loads(p)
        assert numpy.allclose(m, m2)
        assert numpy.allclose(m, m3)
        assert numpy.allclose(m, m4)
        assert not m4.base.is_server
        m4.base._get_in_use_flag_ref().value = 42
        assert m4.base._get_in_use_flag_ref().value == 42
        assert find_numpy_shared_by_shmid(m4.base.mem.shmid)._get_in_use_flag_ref().value == 42
        assert numpy.allclose(m, m4)
        ss = list([find_numpy_shared_by_shmid(_m.base.mem.shmid) for _m in (m2, m3, m4)])
        _m = None
        m2 = m3 = m4 = None
        gc.collect()
        for s in ss:
            assert isinstance(s, SharedNumpyArray)
            assert s.is_server
            assert not s.is_in_use()



from __future__ import annotations
import gc
import sys
import time
import numpy
import threading
import typing
from .basic import Dataset
from returnn.log import log
from returnn.util import NumbersDict


class CachedDataset(Dataset):
    def __init__(self, cache_byte_size=0, **kwargs):
        """
        :param int cache_byte_size:
        """
        super(CachedDataset, self).__init__(**kwargs)
        self.cache_byte_size_total_limit = cache_byte_size
        if cache_byte_size == -1:
            self.cache_byte_size_limit_at_start = 1024**4
        elif cache_byte_size == 0:
            self.cache_byte_size_limit_at_start = 0
        else:
            self.cache_byte_size_limit_at_start = max(cache_byte_size * 2 // 3, 1)
            self.cache_byte_size_total_limit = max(cache_byte_size - self.cache_byte_size_limit_at_start, 1)
        self.num_seqs_cached_at_start = 0
        self.cached_bytes_at_start = 0
        self.nbytes = 0
        self.start_cache_initialized = False
        self.definite_cache_leftover = 0
        self.cache_num_frames_free = 0
        self.preload_set = set([])
        self.preload_end = 0
        self.alloc_intervals = None  # type: typing.Optional[list]
        self._seq_start = []  # [numpy.array([0,0])]  # uses sorted seq idx, see set_batching()
        self._seq_index = []  # type: typing.List[int]  # Via init_seq_order(). seq_index idx -> hdf seq idx
        self._seq_index_inv = {}  # type: typing.Dict[int,int]  # Via init_seq_order(). hdf seq idx -> seq_index idx
        self._index_map = range(len(self._seq_index))  # sorted seq idx -> seq_index idx
        self._tag_idx = {}  # type: typing.Dict[str,int]  # map of tag -> real-seq-idx. call _update_tag_idx
        self.targets = {}
        self.target_keys = (
            []
        )  # the keys for which we provide data; we may have labels for additional keys in self.labels
        self.timestamps = None

    def initialize(self):
        """
        Initialization.
        """
        super(CachedDataset, self).initialize()

        if self.cache_byte_size_limit_at_start > 0:
            self.nbytes = numpy.array([], dtype=numpy.float32).itemsize * (self.num_inputs * self.window + 1 + 1)

            # Calculate cache sizes.
            temp_cache_size_bytes = max(0, self.cache_byte_size_total_limit)
            self.definite_cache_leftover = (
                temp_cache_size_bytes if self.num_seqs_cached_at_start == self.num_seqs else 0
            )
            self.cache_num_frames_free = temp_cache_size_bytes // self.nbytes

            print(
                "cached %i seqs" % self.num_seqs_cached_at_start,
                "%s GB" % (self.cached_bytes_at_start / float(1024 * 1024 * 1024)),
                ("(fully loaded, %s GB left over)" if self.definite_cache_leftover else "(%s GB free)")
                % max(temp_cache_size_bytes / float(1024 * 1024 * 1024), 0),
                file=log.v4,
            )

    def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
        """
        :type epoch: int|None
        :param list[str]|None seq_list: List of sequence tags, to set a predefined order.
        :param list[int]|None seq_order: List of corpus sequence indices, to set a predefined order.
        Initialize lists:
          self.seq_index  # sorted seq idx
        """
        super(CachedDataset, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
        if seq_order is not None:
            seq_index = seq_order
        elif seq_list is not None:
            self._update_tag_idx()
            seq_index = [self._tag_idx[tag] for tag in seq_list]
        else:
            seq_index = self.get_seq_order_for_epoch(
                epoch, self._num_seqs, lambda s: self._get_seq_length_by_real_idx(s)[0]
            )

        old_index_map = self._index_map[:]
        self._index_map = range(len(seq_index))  # sorted seq idx -> seq_index idx

        if isinstance(seq_index, numpy.ndarray) or isinstance(self._seq_index, numpy.ndarray):
            seq_index_unchanged = numpy.array_equal(self._seq_index, seq_index)
        else:
            seq_index_unchanged = self._seq_index == seq_index

        if seq_index_unchanged and self.start_cache_initialized:
            return False

        if epoch is not None:
            # Give some hint to the user in case he is wondering why the cache is reloading.
            print("Reinitialize dataset seq order for epoch %i." % epoch, file=log.v4)

        if (
            self.cache_byte_size_limit_at_start == 0
            or self.num_seqs_cached_at_start != len(seq_index)
            or not self.start_cache_initialized
        ):
            self._seq_index = seq_index
            self._seq_index_inv = {}  # reset, create later if needed
            self._init_seq_starts()
            self._init_alloc_intervals()
            self._init_start_cache()
            self.start_cache_initialized = True
        else:
            if not self._seq_index_inv:
                self._seq_index_inv = dict(
                    zip(self._seq_index, range(len(self._seq_index)))
                )  # hdf seq idx -> seq_index idx
            self._index_map = [self._seq_index_inv[i] for i in seq_index]  # sorted seq idx -> seq_index idx
            if self._index_map == old_index_map:
                return False
        return True

    def get_current_seq_order(self):
        assert self.cache_byte_size_limit_at_start == 0  # not implemented otherwise, we ignore _index_map
        return self._seq_index

    def _get_tag_by_real_idx(self, real_idx):
        raise NotImplementedError

    def _update_tag_idx(self):
        if self._tag_idx:
            return
        for i in range(self._num_seqs):
            self._tag_idx[self._get_tag_by_real_idx(i)] = i

    def batch_set_generator_cache_whole_epoch(self):
        return True

    def _init_alloc_intervals(self):
        if self.cache_byte_size_limit_at_start == 0:
            return
        assert self.num_seqs > 0
        assert self.num_inputs > 0
        assert self.window > 0
        self.preload_set = set([])
        self.alloc_intervals = [
            (0, 0, numpy.zeros([1] + self.get_data_shape("data"), dtype=self.get_data_dtype("data"))),
            (
                self.num_seqs,
                self.num_seqs,
                numpy.zeros([1] + self.get_data_shape("data"), dtype=self.get_data_dtype("data")),
            ),
        ]
        # self.alloc_intervals[i] is (idx start, idx end, data), where
        # idx start/end is the sorted seq idx start/end, end exclusive,
        # and data is a numpy.array.

    def _init_seq_starts(self):
        if self.cache_byte_size_limit_at_start == 0:
            return
        self._seq_start = [self._seq_start[0] * 0]  # idx like in seq_index, *not* real idx
        for i in range(self.num_seqs):
            ids = self._seq_index[i]
            self._seq_start.append(self._seq_start[-1] + self._get_seq_length_by_real_idx(ids))

    def _init_start_cache(self):
        if self.cache_byte_size_limit_at_start == 0:
            return
        if not self.alloc_intervals:
            return
        if not self.nbytes:
            return

        num_cached = 0
        cached_bytes = 0
        for i in range(self.num_seqs):
            if i == num_cached:
                nbytes = self.get_seq_length_nd(i)[0] * self.nbytes
                if self.cache_byte_size_limit_at_start >= cached_bytes + nbytes:
                    num_cached = i + 1
                    cached_bytes += nbytes

        self.num_seqs_cached_at_start = num_cached
        self.cached_bytes_at_start = cached_bytes
        if num_cached > 0:
            self.preload_end = num_cached
            if sys.version_info >= (3, 0):
                threading.Thread(target=self._preload_seqs, args=(0, num_cached), daemon=True).start()
            else:
                threading.Thread(target=self._preload_seqs, args=(0, num_cached)).start()

    def load_seqs(self, start, end):
        """
        Load data sequences.
        As a side effect, will modify / fill-up:
          self.alloc_intervals
          self.targets
        This does some extra logic for the cache and calls self._load_seqs()
        for the real loading.

        :param int start: start sorted seq idx
        :param int end: end sorted seq idx
        """
        assert start >= 0
        assert start <= end

        if self.is_cached(start, end, blocking=True):
            return

        if self.cache_byte_size_limit_at_start > 0:  # If the cache is enabled.
            self._load_seqs_with_cache(start, end)
            return self.is_cached(start, end, blocking=True)

        super(CachedDataset, self).load_seqs(start, end)

    def _load_seqs(self, start, end):
        raise NotImplementedError

    def _load_seqs_with_cache(self, start, end, clear=True):
        if not clear:
            # only remove as many frames as required
            num_needed_cache_frames = self.get_seq_start(end)[0] - self.get_seq_start(start)[0]
            if self.cache_num_frames_free < num_needed_cache_frames:
                self.cache_num_frames_free += self.delete(num_needed_cache_frames - self.cache_num_frames_free)
                gc.collect()
            self.cache_num_frames_free -= num_needed_cache_frames
            threading.Thread(target=self._preload_seqs, args=(start, end)).start()
        else:
            # First, delete everything.
            self.cache_num_frames_free += self.delete(None)
            gc.collect()
            # Preload as much as we can so that we fill up the cache.
            while end < self.num_seqs:
                num_needed_cache_frames = self.get_seq_length_nd(end)[0]
                if self.cache_num_frames_free - num_needed_cache_frames < 0:
                    break
                self.cache_num_frames_free -= num_needed_cache_frames
                end += 1
            self.preload_end = end
            threading.Thread(target=self._preload_seqs, args=(start, end)).start()

    def _preload_seqs(self, start, end):
        print("Preloading cache from", start, "to", end, file=log.v4)
        super(CachedDataset, self).load_seqs(start, end)
        self.preload_end = self.num_seqs_cached_at_start

    def _shuffle_frames_in_seqs(self, start, end):
        """
        :type start: int
        :type end: int
        """
        assert start < end
        assert self.is_cached(start, end)
        alloc_idx = self.alloc_interval_index(start)
        alloc_start, alloc_end, alloc_data = self.alloc_intervals[alloc_idx]
        assert start >= alloc_start
        assert end <= alloc_end
        rnd = numpy.random.RandomState(start)  # Some deterministic way to shuffle!
        num_frames = self._seq_start[end][0] - self._seq_start[start][0]
        assert num_frames > 0
        perm = rnd.permutation(num_frames)
        alloc_offset = self._seq_start[start][0] - self._seq_start[alloc_start][0]
        assert alloc_offset + num_frames <= alloc_data.shape[0]
        # Permute alloc_data.
        data = alloc_data[alloc_offset : alloc_offset + num_frames]
        alloc_data[alloc_offset : alloc_offset + num_frames] = data[perm]
        # Permute targets.
        for k in self.targets:
            idx = self.target_keys.index(k) + 1
            targets = self.targets[k][self._seq_start[idx] : self._seq_start[start][idx] + num_frames]
            self.targets[k][
                self._seq_start[start][idx] : self._seq_start[start][idx]
                + self._seq_start[end][idx]
                - self._seq_start[start][idx]
            ] = targets[perm]

    def _set_alloc_intervals_data(self, idc, data):
        """
        :param int idc: index of sorted seq idx
        :param numpy.ndarray data: raw data
        """
        idi = self.alloc_interval_index(idc)
        assert idi >= 0
        o = self._seq_start[idc][0] - self._seq_start[self.alloc_intervals[idi][0]][0]
        l = data.shape[0]
        x = data
        if self.window > 1:
            x = self._sliding_window(x)
        self.alloc_intervals[idi][2][o : o + l] = x

    def alloc_interval_index(self, ids):
        """
        :param int ids: sorted seq idx
        :return index in self.alloc_intervals
        :rtype: int
        """
        s = 0
        e = len(self.alloc_intervals)
        # Binary search.
        while s < e:
            i = (s + e) // 2
            alloc_start, alloc_end, _ = self.alloc_intervals[i]
            if alloc_start <= ids < alloc_end:
                return i
            elif alloc_start <= ids and ids >= alloc_end:
                if s == i:
                    return -1
                s = i
            elif alloc_start > ids:
                if e == i:
                    return -1
                e = i
            else:
                assert False
        return -1

    def _insert_alloc_interval(self, pos, value, merge=False):
        """
        Insert np.zeros into self.alloc_intervals.
        :param int pos: idx in self.alloc_intervals
        :param (int,int) value: (start,end) like in load_seqs(), sorted seq idx
        :rtype: int
        """
        if value[0] == value[1]:
            return 0
        ci = self.alloc_intervals[pos][1]
        ni = self.alloc_intervals[pos + 1][0]
        xc = self.alloc_intervals[pos][2]
        xn = self.alloc_intervals[pos + 1][2]
        if value[0] == ci and value[1] == ni and merge:
            nj = self.alloc_intervals[pos][0]
            nk = self.alloc_intervals[pos + 1][1]
            del self.alloc_intervals[pos]
            del self.alloc_intervals[pos]
            self.alloc_intervals.insert(
                pos,
                (
                    nj,
                    nk,
                    numpy.concatenate(
                        [
                            xc,
                            numpy.zeros(
                                [self._seq_start[ni][0]] + self.get_data_shape("data"),
                                dtype=self.get_data_dtype("data"),
                            ),
                            xn,
                        ]
                    ),
                ),
            )
            return 0
        elif value[0] == ci and merge:
            nj = self.alloc_intervals[pos][0]
            del self.alloc_intervals[pos]
            self.alloc_intervals.insert(
                pos,
                (
                    nj,
                    value[1],
                    numpy.concatenate(
                        [
                            xc,
                            numpy.zeros(
                                [self._seq_start[value[1]][0] - self._seq_start[ci][0]] + self.get_data_shape("data"),
                                dtype=self.get_data_dtype("data"),
                            ),
                        ]
                    ),
                ),
            )
            return 0
        elif value[1] == ni and merge:
            nk = self.alloc_intervals[pos + 1][1]
            del self.alloc_intervals[pos + 1]
            self.alloc_intervals.insert(
                pos + 1,
                (
                    value[0],
                    nk,
                    numpy.concatenate(
                        [
                            numpy.zeros(
                                [self._seq_start[ni][0] - self._seq_start[value[0]][0]] + self.get_data_shape("data"),
                                dtype=self.get_data_dtype("data"),
                            ),
                            xc,
                        ]
                    ),
                ),
            )
            return 0
        else:
            self.alloc_intervals.insert(
                pos + 1,
                value
                + (
                    numpy.zeros(
                        [self._seq_start[value[1]][0] - self._seq_start[value[0]][0]] + self.get_data_shape("data"),
                        dtype=self.get_data_dtype("data"),
                    ),
                ),
            )
            return 1

    def _remove_alloc_interval(self, pos, value):
        """
        Remove data from self.alloc_intervals.
        :param int pos: idx in self.alloc_intervals
        :param (int,int) value: (start,end) like in load_seqs(), sorted seq idx
        :rtype: int
        """
        ci, ni, xi = self.alloc_intervals[pos]
        if value[0] == ci and value[1] == ni:
            del self.alloc_intervals[pos]
            return -1
        elif value[0] == ci:
            self.alloc_intervals.insert(
                pos, (value[1], ni, xi[self._seq_start[value[1]][0] - self._seq_start[ci][0] :])
            )
            del self.alloc_intervals[pos + 1]
            return 0
        elif value[1] == ni:
            self.alloc_intervals.insert(
                pos, (ci, value[0], xi[: self._seq_start[value[0]][0] - self._seq_start[ci][0]])
            )
            del self.alloc_intervals[pos + 1]
            return 0
        else:
            self.alloc_intervals.insert(
                pos, (value[1], ni, xi[self._seq_start[value[1]][0] - self._seq_start[ci][0] :])
            )
            self.alloc_intervals.insert(
                pos, (ci, value[0], xi[: self._seq_start[value[0]][0] - self._seq_start[ci][0]])
            )
            del self.alloc_intervals[pos + 2]
            return 1

    def _modify_alloc_intervals(self, start, end, invert):
        """
        Inserts/removes sorted seq idx range (start,end).
        :param int start: like in load_seqs(), sorted seq idx
        :param int end: like in load_seqs(), sorted seq idx
        :param bool invert: True->insert, False->remove
        :rtype: list[int]
        :return selection list, modified sorted seq idx in self.alloc_intervals
        """
        if end is None:
            end = start + 1
        if start == end:
            return
        assert start < end
        i = 0
        selection = []  # type: typing.List[int]
        modify = self._insert_alloc_interval if invert else self._remove_alloc_interval
        while i < len(self.alloc_intervals) - invert:
            ni = self.alloc_intervals[i + invert][1 - invert]  # insert mode: start idx of next alloc
            ci = self.alloc_intervals[i][invert]  # insert mode: end idx of cur alloc
            assert ci <= ni
            flag = ((ci <= start < ni), (ci < end <= ni), (ci < start and ni <= start) or (ci >= end and ni > end))
            if not flag[0] and not flag[1]:
                if not flag[2]:
                    selection.extend(range(ci, ni))
                    i += modify(i, (ci, ni))
            elif flag[1]:
                v = (start if flag[0] else ci, end)
                selection.extend(range(v[0], v[1]))
                i += modify(i, v)
                break
            elif flag[0]:
                selection.extend(range(start, ni))
                i += modify(i, (start, ni))
            i += 1
        if self.alloc_intervals[0][0] != 0:
            self.alloc_intervals.insert(
                0, (0, 0, numpy.zeros([1] + self.get_data_shape("data"), dtype=self.get_data_dtype("data")))
            )
        if self.alloc_intervals[-1][1] != self.num_seqs:
            self.alloc_intervals.append(
                (
                    self.num_seqs,
                    self.num_seqs,
                    numpy.zeros([1] + self.get_data_shape("data"), dtype=self.get_data_dtype("data")),
                )
            )
        return selection

    def insert_alloc_interval(self, start, end=None):
        return self._modify_alloc_intervals(start, end, True)

    def remove_alloc_interval(self, start, end=None):
        return self._modify_alloc_intervals(start, end, False)

    def delete(self, nframes):
        """
        :param int|None nframes: how much frames to delete max.
          Note that this limit is not strict. We can end up
          deleting more than nframes.
        :return: number of frames deleted
        :rtype: int
        """
        if nframes is not None:
            if nframes == 0:
                return 0
            assert nframes > 0
        deleted = 0
        i = 0
        while (not nframes or deleted < nframes) and i < len(self.alloc_intervals):
            ai = self.alloc_intervals[i]
            if ai[1] > self.num_seqs_cached_at_start and ai[0] < ai[1]:
                removed = self.remove_alloc_interval(max(ai[0], self.num_seqs_cached_at_start), ai[1])
                self.preload_set -= set(removed)
                deleted += sum([self._get_seq_length_by_real_idx(self._seq_index[i])[0] for i in removed])
            else:
                i += 1
        return deleted

    @property
    def num_seqs(self):
        return len(self._index_map)

    def is_cached(self, start, end, blocking=False):
        """
        :param int start: like in load_seqs(), sorted seq idx
        :param int end: like in load_seqs(), sorted seq idx
        :rtype: bool
        :returns whether we have the full range (start,end) of sorted seq idx
          cached in self.alloc_intervals (end is exclusive).
        """
        if self.cache_byte_size_total_limit == 0:  # disabled cache
            return False
        if start == end:
            return True  # Empty.
        assert start < end
        if blocking and end <= self.preload_end:
            while not set(range(start, end)) <= self.preload_set:
                time.sleep(0.2)
            return True
        return set(range(start, end)) <= self.preload_set

    def _get_seq_length_by_real_idx(self, real_seq_idx):
        """
        :param int real_seq_idx:
        :returns length of the sequence with index 'real_seq_idx'
        :rtype: numpy.ndarray
        """
        raise NotImplementedError

    def get_seq_length_nd(self, sorted_seq_idx):
        """
        :type sorted_seq_idx: int
        :rtype: numpy.ndarray
        """
        real_seq_idx = self._seq_index[self._index_map[sorted_seq_idx]]
        return self._get_seq_length_by_real_idx(real_seq_idx)

    def get_seq_length(self, seq_idx):
        """
        :rtype: NumbersDict
        """
        lengths = self.get_seq_length_nd(seq_idx)
        d = {}
        first_target_idx = 0
        # We allow using only targets. In this case self.num_inputs == 0 and the "data" key is not used.
        if self.num_inputs > 0:
            d["data"] = lengths[0]
            first_target_idx = 1
        for k, l in zip(self.target_keys, lengths[first_target_idx:]):
            d[k] = l
        return NumbersDict(d)

    def get_seq_start(self, sorted_seq_idx):
        """
        :type sorted_seq_idx: int
        :rtype: (int,int)
        """
        return self._seq_start[sorted_seq_idx]

    def get_times(self, sorted_seq_idx):
        seq_start = self.get_seq_start(sorted_seq_idx)[0]
        seq_len = self.get_seq_length_nd(sorted_seq_idx)[0]
        return self.timestamps[seq_start : seq_start + seq_len]

    def get_input_data(self, sorted_seq_idx):
        seq_idx = self._index_map[sorted_seq_idx]
        idi = self.alloc_interval_index(seq_idx)
        assert idi >= 0, "failed to get data for seq %i" % sorted_seq_idx
        alloc_start_seq, alloc_end_seq, alloc_data = self.alloc_intervals[idi]
        o = self.get_seq_start(seq_idx)[0] - self.get_seq_start(alloc_start_seq)[0]
        assert o >= 0
        l = self.get_seq_length_nd(sorted_seq_idx)[0]
        assert alloc_data.shape[0] >= o + l
        return alloc_data[o : o + l]

    def get_data_dim(self, key):
        if key == "data" and self.num_inputs > 0:  # if num_inputs == 0, we allow "data" as a target key
            return self.num_inputs * self.window
        return self.num_outputs[key][0]

    def get_targets(self, target, sorted_seq_idx):
        seq_idx = self._index_map[sorted_seq_idx]
        idx = self.target_keys.index(target) + 1
        seq_start = self.get_seq_start(seq_idx)[idx]
        seq_len = self.get_seq_length_nd(sorted_seq_idx)[idx]
        return self.targets[target][seq_start : seq_start + seq_len]

    def get_target_list(self):
        return list(self.targets.keys())

    def get_tag(self, sorted_seq_idx):
        raise NotImplementedError

    def have_corpus_seq_idx(self):
        return True

    def get_corpus_seq_idx(self, seq_idx):
        """
        :param int seq_idx: sorted sequence index from the current epoch, depending on seq_ordering
        :return: the sequence index as-is in the original corpus. only defined if self.have_corpus_seq_idx()
        :rtype: int
        """
        return self._seq_index[self._index_map[seq_idx]]



"""
Like :mod:`returnn.frontend_api` / :class:`returnn.frontend_api.Frontend`,
but this is for RETURNN internal usage.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar, Any, Optional, Union, Type, Sequence, Tuple
import contextlib


if TYPE_CHECKING:
    from returnn.tensor import Tensor


T = TypeVar("T")  # tf.Tensor, torch.Tensor or so


class InternalFrontend(Generic[T]):
    """
    Private API, most of them operating on raw tensors
    """

    @staticmethod
    def get_dtype_name_raw(raw_tensor: T) -> str:
        """
        :return: dtype of raw tensor, as string
        """
        raise NotImplementedError

    @staticmethod
    def get_ndim_raw(raw_tensor: T) -> int:
        """
        :return: ndim of raw tensor. assumes it is known
        """
        raise NotImplementedError

    @staticmethod
    def get_shape_raw(raw_tensor: T) -> T:
        """
        :return: shape of raw tensor
        """
        raise NotImplementedError

    @staticmethod
    def get_shape_tuple_raw(raw_tensor: T) -> Tuple[Union[int, T]]:
        """
        :return: shape of raw tensor. assumes that ndim is known.
            In eager frameworks, all dims are int.
        """
        raise NotImplementedError

    @staticmethod
    def get_known_shape_raw(raw_tensor: T) -> Tuple[Optional[int]]:
        """
        :return: shape of raw tensor, int for static known, None otherwise. assumes that ndim is known.
            This will not create any ops.
            In eager frameworks, all dims are known.
        """
        raise NotImplementedError

    @staticmethod
    def set_known_shape_raw(raw_tensor: T, shape: Tuple[Optional[int]]) -> None:
        """
        Sets the known shape of the raw tensor.
        This is only supported in graph-based frameworks,
        and just performs a check in eager frameworks.
        """
        # Default implementation for eager-based frameworks.
        assert all(dim is not None for dim in shape)
        rif = get_internal_frontend_by_tensor_type(type(raw_tensor))
        existing_shape = rif.get_known_shape_raw(raw_tensor)
        assert all(dim is not None for dim in existing_shape)
        assert shape == existing_shape

    @staticmethod
    def fill_raw(shape: Union[Sequence[Union[int, T]], T], value: Union[Any, T]) -> T:
        """
        :param shape: shape
        :param value: scalar value to fill
        :return: raw tensor filled with value everywhere
        """
        raise NotImplementedError

    @staticmethod
    def compare_raw(a: T, kind: str, b: T) -> T:
        """
        :param a:
        :param kind: "equal"|"==", "less"|"<", "less_equal"|"<=", "greater"|">", "greater_equal"|">=", "not_equal"|"!="
        :param b:
        :return: a `kind` b
        """
        raise NotImplementedError

    @staticmethod
    def reshape_raw(raw_tensor: T, shape: Union[Sequence[Union[int, T]], T]) -> T:
        """
        :param raw_tensor: raw tensor
        :param shape: new shape
        :return: reshaped raw tensor
        """
        raise NotImplementedError

    @classmethod
    def squeeze_raw(cls, raw_tensor: T, axes: Sequence[int]) -> T:
        """
        :param raw_tensor: raw tensor
        :param axes: axes to squeeze
        :return: squeezed raw tensor
        """
        # Default implementation using reshape_raw.
        known_shape = cls.get_known_shape_raw(raw_tensor)
        assert all([known_shape[axis] == 1 for axis in axes])
        new_shape = [dim for a, dim in enumerate(cls.get_shape_tuple_raw(raw_tensor)) if a not in axes]
        return cls.reshape_raw(raw_tensor, new_shape)

    @staticmethod
    def transpose_raw(raw_tensor: T, perm: Sequence[int]) -> T:
        """
        :param raw_tensor: raw tensor
        :param perm: permutation
        :return: transposed raw tensor
        """
        raise NotImplementedError

    @staticmethod
    def expand_dims_raw(raw_tensor: T, axis: int) -> T:
        """
        :param raw_tensor:
        :param axis:
        :return: raw tensor with new axis
        """
        raise NotImplementedError

    @staticmethod
    def expand_raw(raw_tensor: T, axis: int, dim: Union[int, T]) -> T:
        """
        :param raw_tensor:
        :param axis: shape[axis] must be 1
        :param dim: the new dim for shape[axis]
        :return: shape[axis] expands to dim.
            in PyTorch or other frameworks which support custom strides,
            this is an efficient view and not a copy.
        """
        raise NotImplementedError

    @staticmethod
    def sequence_mask_raw(lengths: T, *, batch_major: bool = True) -> T:
        """
        Like tf.sequence_mask().

        :param lengths: shape (batch,)
        :param batch_major:
        :return: tensor mask of shape (batch,maxlen) if batch_major else (maxlen,batch) of type bool
        """
        raise NotImplementedError

    @staticmethod
    @contextlib.contextmanager
    def name_scope_raw(name: str) -> Any:
        """
        Default implementation for eager-based frameworks:
        Do nothing, tensors do not have a name.

        :param name:
        :return: context manager
        """
        # Default implementation for eager-based frameworks
        pass  # nothing to do

    @staticmethod
    @contextlib.contextmanager
    def control_dependencies_raw(dependencies: Sequence[Any]) -> Any:
        """
        Default implementation for eager-based frameworks:
        Do nothing, we expect that the dependencies are already executed.

        :param dependencies: raw tensors or ops
        :return: context manager
        """
        # Default implementation for eager-based frameworks
        yield

    @staticmethod
    def identity_with_control_dependencies_raw(raw_tensor: T, dependencies: Sequence[Any]) -> T:
        """
        Default implementation for eager-based frameworks:
        Do nothing, we expect that the dependencies are already executed.

        :param raw_tensor: raw tensor
        :param dependencies: raw tensors or ops
        :return: raw tensor
        """
        # Default implementation for eager-based frameworks
        return raw_tensor

    @staticmethod
    def create_placeholder(tensor: Tensor) -> T:
        """
        :return: tf.placeholder in TF

        This is really only for TensorFlow for the deprecated option auto_create_placeholders
        and should not be used in other backends,
        even in graph-based backends.
        Rather, the logic to create placeholders should be done elsewhere.
        """
        raise Exception("create_placeholder not supported by backend")

    @staticmethod
    def runtime_sanity_checks(tensor: Tensor) -> Any:
        """
        Checks whether the tensor.raw_tensor is consistent with the tensor metadata.

        In graph-based frameworks (TF graph), we return some operation here.
        In eager frameworks, we would not return anything but instead directly perform the checks.
        """
        # By default, we do not do any checks. This is optional for the backend.
        pass

    @staticmethod
    def is_valid_in_current_graph(tensor: Tensor) -> bool:
        """
        :return: whether the raw tensor is valid in the current graph.
            In eager-mode frameworks, this is always true -- there is no graph.
        """
        return True

    @staticmethod
    def format_graph_output(raw_tensor: T, *, max_depth: Optional[int] = None) -> str:
        """
        :return: the computation graph leading to this tensor formatted.
            In eager-mode frameworks, this is not supported and returns None.
        """
        return "<no-graph>"


def get_internal_frontend_by_tensor_type(tensor_type: Type[T]) -> Type[InternalFrontend[T]]:
    """
    :param tensor_type: e.g. tf.Tensor
    :return: internal frontend class
    """
    from .frontend_api import get_frontend_by_tensor_type

    rf = get_frontend_by_tensor_type(tensor_type)
    # noinspection PyProtectedMember
    return rf._internal_frontend
